""" Metafile Lister.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""

import hashlib
import sys
import traceback

from pathlib import Path

import bencode

from pyrosimple.error import EX_DATAERR, EX_SOFTWARE
from pyrosimple.scripts.base import ScriptBase


class MetafileLister(ScriptBase):
    """List contents of a bittorrent metafile."""

    ENABLE_PROGRESS = True

    def add_options(self):
        """Add program options."""
        self.parser.add_argument("metafile", nargs="+", help="Torrent files to display")
        self.add_bool_option(
            "--reveal",
            help="show full announce URL including keys, as well as full piece information",
        )
        self.add_bool_option(
            "--raw", "--json", help="print the metafile's raw content in JSON format"
        )
        self.add_bool_option(
            "-V",
            "--skip-validation",
            help="show broken metafiles with an invalid structure",
        )
        self.add_value_option(
            "-o",
            "--output",
            "KEY1,KEY2.SUBKEY,...",
            action="append",
            default=[],
            help="select fields to print, output is separated by TABs;"
            " note that __file__ is the path to the metafile,"
            " __hash__ is the info hash,"
            " and __size__ is the data size in bytes",
        )
        self.add_value_option(
            "-c",
            "--check-data",
            "PATH",
            help="check the hash against the data in the given path",
        )

    def mainloop(self):
        """The main loop."""
        from pyrosimple.util import metafile  # pylint: disable=import-outside-toplevel

        self.args = self.options.metafile
        if not self.args:
            self.parser.print_help()
            self.parser.error("No metafiles given, nothing to do!")
            self.parser.exit()

        validation_errors = (ValueError, KeyError, bencode.BencodeDecodeError)

        for idx, filename in enumerate(self.args):
            if idx and not self.options.output and not self.options.raw:
                print()
                print("~" * 79)

            try:
                # Read and check metafile
                try:
                    filename = Path(filename)
                    torrent = metafile.Metafile.from_file(filename)
                    if not self.options.skip_validation:
                        try:
                            torrent.check_meta()
                        except validation_errors:
                            self.return_code = EX_SOFTWARE
                            raise
                except OSError as exc:
                    self.fatal(
                        "Can't read '%s' (%s)"
                        % (
                            filename,
                            str(exc).replace(": '%s'" % filename, ""),
                        )
                    )
                    raise

                display_nothing = bool(self.options.output == [""])  # i.e `-o ""`

                output_values = torrent.dict_copy()
                if not self.options.reveal:
                    # Shorten useless binary piece hashes
                    if "info" in output_values:
                        count = len(output_values["info"]["pieces"]) / len(
                            hashlib.sha1().digest()
                        )
                        output_values["info"]["pieces"] = f"<{count} piece hashes>"
                    if "piece layers" in output_values:
                        for layer, pieces in output_values["piece layers"].items():
                            count = len(pieces) / 16 * 1024 * 1024
                            output_values["piece layers"][
                                layer
                            ] = f"<{count} piece hashes>"
                if self.options.output and not display_nothing:
                    output_values = []

                    def splitter(fields):
                        "Yield single names for a list of comma-separated strings."
                        for flist in fields:
                            for field in flist.split(","):
                                yield field.strip()

                    data = {}
                    data["__file__"] = str(filename)
                    if "info" in torrent:
                        data["__hash__"] = torrent.info_hash()
                        data["__size__"] = torrent.data_size()

                    for field in splitter(self.options.output):
                        try:
                            if field in data:
                                val = data[field]
                            else:
                                val = torrent.fetch_field(field)
                                if val is None:
                                    raise KeyError()
                        except KeyError:
                            self.log.error("%s: Field %r not found", filename, field)
                            output_values.append(None)
                        else:
                            output_values.append(val)

                if display_nothing:
                    pass
                elif self.options.raw:
                    from pyrosimple.util.fmt import (  # pylint: disable=import-outside-toplevel
                        BencodeJSONEncoder,
                    )

                    print(BencodeJSONEncoder(indent=2).encode(output_values))
                elif self.options.output:
                    print(
                        "\t".join(
                            [str(o) if o is not None else "" for o in output_values]
                        )
                    )
                else:
                    try:
                        print(
                            "\n".join(torrent.listing(masked=not self.options.reveal))
                        )
                    except validation_errors as exc:
                        self.log.error(
                            "Bad metafile %r (%s: %s)",
                            str(filename),
                            type(exc).__name__,
                            exc,
                        )
                        print(traceback.format_exc(), end="")
                        if not self.options.skip_validation:
                            self.return_code = EX_SOFTWARE
                if self.options.check_data:
                    # pylint: disable=import-outside-toplevel

                    from pyrosimple.util.metafile import PieceFailer
                    from pyrosimple.util.ui import HashProgressBar

                    piece_logger = PieceFailer(torrent, self.log)
                    try:
                        if self.options.progress == "on":
                            with HashProgressBar() as pb:
                                progress_callback = pb().progress_callback

                                torrent.hash_check(
                                    Path(self.options.check_data),
                                    progress_callback=progress_callback,
                                    piece_callback=piece_logger.check_piece,
                                )
                        else:
                            torrent.hash_check(
                                Path(self.options.check_data),
                                piece_callback=piece_logger.check_piece,
                            )
                    except OSError as exc:
                        print(
                            f"ERROR: File {str(filename)!r} did not hash check: {exc}"
                        )
                        sys.exit(EX_DATAERR)
            except validation_errors as exc:
                self.log.error(
                    "Bad metafile %r (%s: %s)", str(filename), type(exc).__name__, exc
                )
                print(traceback.format_exc(), end="")
                if not self.options.skip_validation:
                    self.return_code = EX_SOFTWARE


def run():  # pragma: no cover
    """The entry point."""
    MetafileLister().run()


if __name__ == "__main__":
    run()



"""Data Formatting.

Copyright (c) 2009, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import datetime
import json
import logging
import os
import re
import shlex
import time

from pathlib import Path
from pprint import pformat
from typing import Optional

from pyrosimple import torrent
from pyrosimple.util import pymagic


log = logging.getLogger(__name__)


def human_size(size: float) -> str:
    """Return a human-readable representation of a byte size.

    @param size: Number of bytes as an integer or string.
    @return: String of length 10 with the formatted result.
    """
    if isinstance(size, str):
        size = float(size, 10)

    if size < 0:
        return "-??? bytes"

    if size < 1024:
        return f"{int(size):4d} bytes".lstrip()

    rem = float(size)
    for unit in ("KiB", "MiB", "GiB", "TiB", "PiB"):
        rem /= 1024.0
        if rem < 1024:
            return f"{rem:6.1f} {unit}".lstrip()

    return f"{rem:6.1f} PiB".lstrip()


def fmt_shell(string: str) -> str:
    """Quote a string for use in shell scripts."""
    return shlex.quote(string)


def fmt_sz(intval: int) -> str:
    """Format a byte sized value."""
    try:
        return human_size(intval).rjust(10)
    except (ValueError, TypeError):
        return "N/A".rjust(10)


def fmt_iso(timestamp: float) -> str:
    """Format a UNIX timestamp to an ISO datetime string."""
    try:
        return iso_datetime(timestamp)
    except (ValueError, TypeError):
        return "N/A".rjust(len(iso_datetime(0)))


def fmt_duration(duration: int) -> str:
    """Format a duration value in seconds to a readable form."""
    try:
        return human_duration(float(duration), 0, 2, True)
    except (ValueError, TypeError):
        return "N/A".rjust(len(human_duration(0, 0, 2, True)))


def fmt_delta(timestamp) -> str:
    """Format a UNIX timestamp to a delta (relative to now)."""
    try:
        return human_duration(float(timestamp), precision=2, short=True)
    except (ValueError, TypeError):
        return "N/A".rjust(len(human_duration(0, precision=2, short=True)))


def fmt_pc(floatval: float):
    """Scale a ratio value to percent."""
    return round(float(floatval) * 100.0, 2)


def fmt_strip(val: str) -> str:
    """Strip leading and trailing whitespace."""
    return str(val).strip()


def fmt_subst(val, regex, subst):
    """Replace regex with string."""
    return re.sub(regex, subst, val)


def fmt_mtime(val: str) -> float:
    """Modification time of a path."""
    p = Path(str(val))
    if p.exists():
        return p.stat().st_mtime
    return 0.0


def fmt_pathbase(val: str) -> str:
    """Base name of a path."""
    return os.path.basename(val or "")


def fmt_pathname(val: str) -> str:
    """Base name of a path, without its extension."""
    return os.path.splitext(os.path.basename(val or ""))[0]


def fmt_raw(val):
    """Show the raw value of the field."""
    # A little magic to allow showing the raw value of a field in rtcontrol
    return val


def fmt_fmt(val, field):
    """Apply a field-specific formatter (if present)"""

    # If val is a RtorrentItem, fetch `field` from it before formatting. This
    # is to allow `d|fmt('is_private')` vs. the redundant `d.is_private|fmt('is_private')`.
    # Be aware that using the former in rtcontrol templates breaks the field auto-detection.
    if field not in torrent.engine.FIELD_REGISTRY:
        return val
    if isinstance(val, torrent.rtorrent.RtorrentItem):
        val = getattr(val, field)
    formatter = torrent.engine.FIELD_REGISTRY[field].formatter
    if formatter:
        return formatter(val)
    return val


def fmt_pathext(val: str) -> str:
    """Extension of a path (including the '.')."""
    return os.path.splitext(val or "")[1]


def fmt_pathdir(val: str):
    """Directory containing the given path."""
    return os.path.dirname(val or "")


def fmt_json(val):
    """JSON serialization."""
    return json.dumps(val, cls=pymagic.JSONEncoder)


def iso_datetime(timestamp: Optional[float] = None) -> str:
    """Convert UNIX timestamp to ISO datetime string.

    @param timestamp: UNIX epoch value (default: the current time).
    @return: Timestamp formatted as "YYYY-mm-dd HH:MM:SS".
    """
    if timestamp is None:
        timestamp = time.time()
    return datetime.datetime.fromtimestamp(timestamp).isoformat(" ")[:19]


def iso_datetime_optional(timestamp) -> str:
    """Convert UNIX timestamp to ISO datetime string, or "never".

    @param timestamp: UNIX epoch value.
    @return: Timestamp formatted as "YYYY-mm-dd HH:MM:SS", or "never" for false values.
    """
    if timestamp:
        return iso_datetime(timestamp)
    return "never"


def bytes_from_human(val: str) -> int:
    """Convert a human string to integer bytes. Follows the same logic
    as rtcontrol's byte filter"""
    units = {"b": 1, "k": 1024, "m": 1024**2, "g": 1024**3}
    lower_val = str(val).lower()
    if any(lower_val.endswith(i) for i in units):
        scale = units[lower_val[-1]]
        val = val[:-1]
    else:
        scale = 1
    return int(val) * scale


def human_duration(
    time1: float, time2: Optional[float] = None, precision: int = 0, short: bool = False
) -> str:
    """Return a human-readable representation of a time delta.

    @param time1: Relative time value.
    @param time2: Time base (C{None} for now; 0 for a duration in C{time1}).
    @param precision: How many time units to return (0 = all).
    @param short: Use abbreviations, and right-justify the result to always the same length.
    @return: Formatted duration.
    """
    if time2 is None:
        time2 = time.time()

    duration = (time1 or 0) - time2
    direction = (
        " ago" if duration < 0 else ("+now" if short else " from now") if time2 else ""
    )
    duration = abs(duration)
    parts = [
        ("weeks", duration // (7 * 86400)),
        ("days", duration // 86400 % 7),
        ("hours", duration // 3600 % 24),
        ("mins", duration // 60 % 60),
        ("secs", duration % 60),
    ]

    # Kill leading zero parts
    while len(parts) > 1 and parts[0][1] == 0:
        parts = parts[1:]

    # Limit to # of parts given by precision
    if precision:
        parts = parts[:precision]

    numfmt = ("%d", "%d"), ("%4d", "%2d")
    fmt = "%1.1s" if short else " %s"
    sep = " " if short else ", "
    result = (
        sep.join(
            (numfmt[bool(short)][bool(idx)] + fmt)
            % (val, key[:-1] if val == 1 else key)
            for idx, (key, val) in enumerate(parts)
            if val  # or (short and precision)
        )
        + direction
    )

    if not time1:
        result = "never" if time2 else "N/A"

    if precision and short:
        return result.rjust(1 + precision * 4 + (4 if time2 else 0))
    return result


def convert_strings_in_iter(obj):
    """Helper function to nicely format results"""
    if isinstance(obj, bytes):
        obj = obj.decode()
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[convert_strings_in_iter(k)] = convert_strings_in_iter(v)
    elif isinstance(obj, list):
        for k, v in enumerate(obj):
            obj[k] = convert_strings_in_iter(v)
    return obj


def rpc_result_to_string(result) -> str:
    """Helper function to nicely format results"""
    result = convert_strings_in_iter(result)

    if isinstance(result, str):
        return result
    if isinstance(result, bytes):
        return result.decode()
    if hasattr(result, "__iter__"):
        return "\n".join(
            i if isinstance(i, str) else pformat(i, width=240) for i in result
        )
    return repr(result)


def keys_string(d):
    """Recursive convert any dictionary keys from bytes to hex strings"""
    rval = {}
    if not isinstance(d, dict):
        if isinstance(d, (tuple, list, set)):
            v = [keys_string(x) for x in d]
            return v
        else:
            return d

    for k, v in d.items():
        if isinstance(k, bytes):
            try:
                k = k.decode()
            except UnicodeDecodeError:
                k = k.hex().upper()
        if isinstance(v, dict):
            v = keys_string(v)
        elif isinstance(v, (tuple, list, set)):
            v = [keys_string(x) for x in v]
        rval[k] = v
    return rval


class BencodeJSONEncoder(json.JSONEncoder):
    """Small helper class to translate bytes"""

    def default(self, o):
        if isinstance(o, bytes):
            return o.hex().upper()
        return super().default(o)

    def encode(self, o):
        if isinstance(o, dict):
            o = keys_string(o)
        return super().encode(o)



""" Metafile Creator.

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""

import argparse
import fnmatch
import hashlib
import os
import re
import time

from pathlib import Path
from typing import Dict
from urllib.parse import parse_qs

import bencode

from pyrosimple import error
from pyrosimple.scripts.base import ScriptBase


class MetafileCreator(ScriptBase):
    """
    Create a bittorrent metafile.
    """

    ENABLE_PROGRESS = True

    def add_options(self):
        """Add program options."""
        super().add_options()
        # pylint: disable=import-outside-toplevel
        from pyrosimple.util.fmt import bytes_from_human

        self.parser.add_argument(
            "dir_or_file", help="Directory or file to create the torrent from"
        )
        self.parser.add_argument(
            "tracker_url_or_alias",
            nargs="*",
            help="A URL or tracker alias to set the announce to.",
        )
        self.add_bool_option("-p", "--private", help="disallow DHT and PEX")
        self.add_bool_option("--no-date", help="leave out creation date")
        self.add_bool_option("--add-padding", help=argparse.SUPPRESS)
        self.add_value_option(
            "-o",
            "--output-filename",
            "PATH",
            help="optional file name (or target directory) for the metafile",
        )
        self.add_value_option(
            "-r",
            "--root-name",
            "NAME",
            help="optional root name (default is basename of the data path)",
        )
        self.add_value_option(
            "-m",
            "--magnet-watch",
            "NAME",
            help="path to place .meta files from magnet links",
        )
        self.add_value_option(
            "-x",
            "--exclude",
            "PATTERN",
            action="append",
            default=[],
            help="exclude files matching a glob pattern from hashing; can be specified multiple times",
        )
        self.add_value_option(
            "--piece-size",
            "SIZE",
            default="0",
            type=bytes_from_human,
            help="specify the piece size manually: allows byte sizes (e.g. 5M)",
        )
        self.add_value_option(
            "--piece-size-min",
            "SIZE",
            default="32K",
            type=bytes_from_human,
            help="specify a minimum piece size",
        )
        self.add_value_option(
            "--piece-size-max",
            "SIZE",
            default="16M",
            type=bytes_from_human,
            help="specify a maximum piece size",
        )
        self.add_value_option(
            "--comment", "TEXT", help="optional human-readable comment"
        )
        self.add_value_option(
            "-s",
            "--set",
            "KEY=VAL",
            action="append",
            default=[],
            help="set a specific key to the given value; omit the '=' to delete a key; can be specified multiple times",
        )
        self.add_bool_option(
            "-H",
            "--hashed",
            "--fast-resume",
            help="create second metafile containing libtorrent fast-resume information",
        )

    # TODO: Optionally pass torrent directly to rTorrent (--load / --start)
    # TODO: Optionally limit disk I/O bandwidth used (incl. a config default!)
    # TODO: Support multi-tracker extension ("announce-list" field)
    # TODO: DHT "nodes" field?! [[str IP, int port], ...]
    # TODO: Web-seeding http://www.getright.com/seedtorrent.html
    #       field 'url-list': ['http://...'] on top-level

    def make_magnet_meta(self, magnet_url):
        """Create a magnet-url torrent."""

        if magnet_url.startswith("magnet:"):
            magnet_url = magnet_url[7:]
        meta = {"magnet-uri": "magnet:" + magnet_url}
        magnet_params = parse_qs(magnet_url.lstrip("?"))

        meta_name = magnet_params.get(
            "xt", [hashlib.sha1(magnet_url.encode()).hexdigest()]
        )[0]
        if "dn" in magnet_params:
            meta_name = f"{magnet_params['dn'][0]}-{meta_name}"
        meta_name = (
            re.sub(r"[^-_,a-zA-Z0-9]+", ".", meta_name)
            .strip(".")
            .replace("urn.btih.", "")
        )

        if not self.options.magnet_watch:
            self.fatal("You MUST set the '--magnet-watch' config option!")
        meta_path = os.path.join(
            self.options.magnet_watch, f"magnet-{meta_name}.torrent"
        )
        self.log.debug("Writing magnet-uri metafile %r", meta_path)

        try:
            bencode.bwrite(meta, meta_path)
        except OSError as exc:
            self.fatal("Error writing magnet-uri metafile %r (%s)", (meta_path, exc))
            raise

    def from_path(self, datapath, **kwargs):
        """Helper wrapper for the metafile method, to avoid repeating code"""
        from pyrosimple import config  # pylint: disable=import-outside-toplevel
        from pyrosimple.util import metafile  # pylint: disable=import-outside-toplevel

        return metafile.Metafile.from_path(
            datapath,
            self.args[1],
            root_name=self.options.root_name,
            private=self.options.private,
            created_by="PyroSimple",
            piece_size=self.options.piece_size,
            piece_size_min=self.options.piece_size_min,
            piece_size_max=self.options.piece_size_max,
            add_padding=self.options.add_padding,
            ignore=[
                re.compile(fnmatch.translate(glob))
                for glob in self.options.exclude + config.settings.MKTOR_IGNORE
            ],
            **kwargs,
        )

    def mainloop(self):
        """The main loop."""
        from pyrosimple import config  # pylint: disable=import-outside-toplevel

        if "=urn:btih:" in self.options.dir_or_file:
            # Handle magnet link
            self.make_magnet_meta(self.options.dir_or_file)
            return
        self.args = [self.options.dir_or_file] + self.options.tracker_url_or_alias

        if not self.args:
            self.parser.print_help()
            self.parser.error("No arguments given, nothing to do!")
            self.parser.exit()
        elif len(self.args) < 2:
            self.parser.error(
                "Expected a path and at least one announce URL, got: %s"
                % (" ".join(self.args),)
            )

        # Validate tracker list
        tracker_urls: Dict[str, str] = {}
        for tracker_url in self.args[1:]:
            found_url = ""
            try:
                tracker_alias, found_urls = config.lookup_announce_url(tracker_url)
                found_url = found_urls[0]
            except KeyError:
                tracker_alias = config.map_announce2alias(tracker_url)
                found_url = tracker_url
            if not found_url:
                raise error.ConfigurationError(
                    f"Announce '{tracker_url}' is not a valid URL or alias"
                )
            tracker_urls[tracker_alias] = found_url

        # Create and configure metafile factory
        datapath = Path(self.args[0])
        metapath = Path(datapath)
        if self.options.output_filename:
            metapath = Path(self.options.output_filename)
            if metapath.is_dir():
                metapath = metapath.joinpath(datapath.name).with_suffix(".torrent")
        else:
            if metapath.is_dir():
                metapath = Path(str(datapath) + ".torrent")
            else:
                metapath = datapath.with_suffix(".torrent")

        # Build progress bar
        # pylint: disable=import-outside-toplevel
        from pyrosimple.util.ui import HashProgressBar

        if self.options.progress == "on":
            with HashProgressBar() as pb:
                c = pb()

                def pb_tracker(totalhashed, totalsize):
                    c.total = totalsize
                    c.items_completed = totalhashed
                    c.progress_bar.invalidate()

                torrent = self.from_path(
                    datapath,
                    progress=pb_tracker,
                )
        else:
            torrent = self.from_path(
                datapath,
            )
        torrent["created by"] = "PyroSimple"
        if self.options.comment:
            torrent["comment"] = self.options.comment
        if not self.options.no_date:
            torrent["creation date"] = int(time.time())

        for alias, announce in tracker_urls.items():
            torrent["announce"] = announce
            torrent["info"]["source"] = alias
            torrent["info"]["x_cross_seed"] = hashlib.md5(announce.encode()).hexdigest()
            torrent.assign_fields(self.options.set)
            if len(self.args) == 2:
                # If only one announce, just save to file
                save_metapath = metapath
            else:
                save_metapath = Path(f"{alias}_{metapath}")
            self.log.info("Writing metafile %s", save_metapath)
            torrent.save(save_metapath)

            # Create second metafile with fast-resume?
            if self.options.hashed:
                try:
                    torrent.add_fast_resume(datapath)
                except OSError as exc:
                    self.fatal(f"Error making fast-resume data ({exc})")
                    raise

                hashed_path = Path(
                    re.sub(r"\.torrent$", "", str(save_metapath)) + "-resume.torrent"
                )
                self.log.info("Writing fast-resume metafile %s", hashed_path)
                try:
                    torrent.save(hashed_path)
                except OSError as exc:
                    self.fatal(
                        f"Error writing fast-resume metafile {hashed_path!r} ({exc})"
                    )
                    raise


def run():  # pragma: no cover
    """The entry point."""
    MetafileCreator().run()


if __name__ == "__main__":
    run()



""" rTorrent Queue Manager.

    Copyright (c) 2012 The PyroScope Project <pyroscope.project@gmail.com>
"""

import time

from pyrosimple.job.base import MatchableJob
from pyrosimple.util import matching, rpc


class QueueManager(MatchableJob):
    """rTorrent queue manager implementation."""

    def __init__(self, config=None):
        """Set up queue manager."""
        # Allow usage of old field names
        if "startable" in config and "matcher" not in config:
            config["matcher"] = config["startable"]
        if "sort_fields" in config and "sort" not in config:
            config["sort"] = config["sort_fields"]
        config.setdefault(
            "matcher", 'is_complete=no is_ignored=no is_open=no message="" done=0'
        )
        super().__init__(config)
        self.last_start: float = 0
        self.downloading_count: int = 0
        self.allowed_start_count: int = 0

        self.config.setdefault("viewname", "main")
        self.config.setdefault("start_at_once", 1)
        self.config.setdefault("intermission", 120)
        self.config.setdefault("downloading", "is_active=1 is_complete=0")
        self.config.setdefault("downloading_min", 0)
        self.config.setdefault("downloading_max", 20)
        self.config.setdefault("downloading_traffic_max", 0)
        self.config["downloading"] = matching.create_matcher(self.config["downloading"])
        self.config["log_to_client"] = matching.truth(
            self.config.get("log_to_client", True)
        )
        self.log.info(
            "Startable matcher is: %s",
            self.config["matcher"],
        )
        self.log.info(
            "Downloading matcher is: %s",
            self.config["downloading"],
        )

    def run(self):
        # Check intermission delay
        now = time.monotonic()
        delayed = int(self.last_start + self.config["intermission"] - now)
        if delayed > 0:
            self.log.debug(
                "Skipping start due to %ds intermission with %ds left",
                self.config["intermission"],
                delayed,
            )
            return
        downloading = list(self.engine.view("incomplete", self.config["downloading"]))
        self.downloading_count = len(downloading)
        # Check download traffic
        if self.config["downloading_traffic_max"] > 0:
            down_traffic = sum(i.down for i in downloading)
            self.log.debug(
                "%d downloading, download traffic is %d", len(downloading), down_traffic
            )
            if down_traffic > int(self.config["downloading_traffic_max"]):
                self.log.debug(
                    "Skipping start due to max download traffic '%s' reached",
                    self.config["downloading_traffic_max"],
                )
                return
        self.allowed_start_count = max(
            self.config["start_at_once"],
            self.config["downloading_min"] - len(downloading),
        )
        if self.allowed_start_count <= 0:
            self.log.debug("Skipping start due to allowed start count reached")
        if self.downloading_count >= self.config["downloading_max"]:
            self.log.debug(
                "Skipping start due to maximum downloading torrents %d reached",
                self.config["downloading_max"],
            )
        self.log.debug("Starting torrents (capped at %d)", self.allowed_start_count)
        # Run parent method
        super().run()

    def run_item(self, item):
        if self.allowed_start_count <= 0:
            return
        if self.downloading_count >= self.config["downloading_max"]:
            return
        if self.config["dry_run"]:
            self.log.info(
                "Would start %s '%s'",
                item.hash,
                item.name,
            )
        else:
            self.log.info(
                "Starting %s '%s'",
                item.hash,
                item.name,
            )
            item.start()
            if self.config["log_to_client"]:
                self.engine.open().log(
                    rpc.NOHASH,
                    f"{self.__class__.__name__}: Started '{item.name}' [{item.alias}]",
                )
        self.downloading_count += 1
        self.allowed_start_count -= 1
        self.last_start = time.monotonic()



""" RTorrent client proxy.

    Copyright (c) 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import base64
import functools
import json
import logging
import random
import urllib

from typing import Any, Dict, List, Tuple, cast
from xmlrpc import client as xmlrpclib

from pyrosimple.io import scgi


logger = logging.getLogger(__name__)

NOHASH = (
    ""  # use named constant to make new-syntax commands with no hash easily searchable
)


class JSONRPCEncoder(json.JSONEncoder):
    """Encode xmlrpc.Binary data in a format jesec/rtorrent
    expects. The necessary command logic when handling load vs
    load.raw still needs to be handled separately."""

    def default(self, o):
        if isinstance(o, xmlrpclib.Binary):
            return "data:base64," + base64.b64encode(o.data).decode("ascii")
        return json.JSONEncoder.default(self, o)


CACHE_METHOD = {
    "system.api_version",
    "system.client_version",
    "system.library_version",
}


class RpcError(xmlrpclib.Fault):
    """Base class for XMLRPC protocol errors."""

    def __init__(self, faultString: str, faultCode: int = -500):
        super().__init__(faultCode, faultString)


class HashNotFound(RpcError):
    """Non-existing or disappeared hash."""

    def __init__(self, faultString: str, faultCode: int = -404):
        super().__init__(faultString, faultCode)


ERRORS = (RpcError,) + scgi.ERRORS


class RTorrentProxy(xmlrpclib.ServerProxy):
    # pylint: disable=super-init-not-called
    """Proxy to rTorrent's RPC interface.

    Method calls are built from attribute accesses, i.e. you can do
    something like `proxy.system.client_version()`.

    All methods from ServerProxy are being overridden due to the
    combination of self.__var name mangling and the
    __call__/__getattr__ magic.
    """

    def __init__(
        self,
        url,
        transport=None,
        encoding=None,
        verbose=False,
        allow_none=False,
        use_datetime=False,
        use_builtin_types=False,
        *,
        headers=(),
        context=None,
    ):
        # Get the url
        parsed_url = urllib.parse.urlsplit(url)
        queries = urllib.parse.parse_qs(parsed_url.query)
        # Config the connection details
        self.__rpc_codec = queries.get("rpc", ["xml"])[0]
        self.__url = url
        self.__host = parsed_url.netloc
        self.__handler = urllib.parse.urlunsplit(["", "", *parsed_url[2:]])

        if transport is None:
            if self.__rpc_codec == "json":
                codec = json
                headers = [("CONTENT_TYPE", "application/json")]
            elif self.__rpc_codec == "xml":
                codec = xmlrpclib
                headers = [("CONTENT_TYPE", "text/xml")]
            else:
                raise ValueError(f"Unknown RPC protocol type {codec}")
            handler = scgi.transport_from_url(url)
            transport = handler(
                url=self.__url,
                use_datetime=use_datetime,
                use_builtin_types=use_builtin_types,
                codec=codec,
                headers=headers,
            )
        self.__transport = transport
        self.__encoding = encoding or "utf-8"
        self.__verbose = verbose
        self.__allow_none = allow_none

    def __close(self):
        self.__transport.close()

    def __request_xml(self, methodname: str, params: Tuple[Any]):
        # Verbatim from parent method
        request = xmlrpclib.dumps(
            params,
            methodname,
            encoding=self.__encoding,
            allow_none=self.__allow_none,
        ).encode(self.__encoding, "xmlcharrefreplace")
        if self.__verbose:
            logger.info("req: %s", request)

        try:
            response = self.__transport.request(
                self.__host, self.__handler, request, verbose=self.__verbose
            )
        except xmlrpclib.Fault as exc:
            raise RpcError(exc.faultString, exc.faultCode) from exc

        if len(response) == 1:
            return response[0]
        return response

    def __batch_request_json(self, calls: List) -> List:
        """Handle JSON-RPC multicalls in place of XMLRPC's built-in
        system.multicall.
        """
        batch_call: List[Dict] = [
            {
                "jsonrpc": "2.0",
                "method": call["methodName"],
                "params": call["params"] or [""],
                "id": index,
            }
            for index, call in enumerate(calls)
        ]
        request: bytes = json.dumps(batch_call).encode(
            self.__encoding, "xmlcharrefreplace"
        )
        response: Tuple[Dict] = cast(
            Tuple[Dict],
            self.__transport.request(
                self.__host,
                self.__handler,
                request,
                verbose=self.__verbose,
            ),
        )

        def sort_key(i: Dict) -> int:
            return int(i["id"])

        if any("error" in r for r in response):
            uniq_errors = {str(r["error"]) for r in response if "error" in r}
            raise RpcError(f"Errors in JSON-RPC batch call: {uniq_errors}")
        result = [[r["result"]] for r in sorted(response, key=sort_key)]
        return result

    def __request_json(self, methodname, params):
        if not params:
            params = [""]

        # system.multicall isn't defined by rTorrent, but by the XMLRPC
        # library, so we intercept it and turn it into a batch request.
        if methodname == "system.multicall":
            return self.__batch_request_json(params[0])

        # This random ID feels silly but there's not much need for
        # anything better at the moment since the RPC interface is
        # synchronous.
        rpc_id = random.randint(0, 100)
        request = (
            JSONRPCEncoder(separators=(",", ":"))
            .encode(
                {
                    "params": params,
                    "method": methodname,
                    "jsonrpc": "2.0",
                    "id": rpc_id,
                }
            )
            .encode(self.__encoding, "xmlcharrefreplace")
        )
        if self.__verbose:
            logger.info("req: %s", request)

        response: Dict = dict(
            self.__transport.request(
                self.__host,
                self.__handler,
                request,
                verbose=self.__verbose,
            )
        )

        if response["id"] != rpc_id:
            raise ValueError(
                f"RPC IDs do not match: sent={rpc_id} received={response['id']}"
            )
        if "error" in response:
            if "message" in response["error"]:
                if (
                    response["error"]["message"]
                    == "invalid parameters: info-hash not found"
                ):
                    raise HashNotFound(str(response["error"]))
                if "code" in response["error"]:
                    raise RpcError(
                        response["error"]["message"], response["error"]["code"]
                    )
            raise RpcError(f"Received error: {response['error']}")
        if "result" not in response:
            raise ValueError(f"Result not found in response: {response}")
        return response["result"]

    def __request(self, methodname, params):
        """Determines whether or not a request should be cached, then
        passes it to the appropriate method

        """
        if methodname in CACHE_METHOD:
            return self.__cached_request(methodname, params)
        return self.__request_switch(methodname, params)

    @functools.lru_cache(maxsize=32)
    def __cached_request(self, methodname, params):
        """Simpled cache pass-through method to more easily take
        advantage of functools.lru_cache

        """
        return self.__request_switch(methodname, params)

    def __request_switch(self, methodname, params):
        """Determines whether the request should go through XMLRPC or
        JSON-RPC and calls the appropriate method.

        """
        logger.debug("method '%s', params %s", methodname, params)
        try:
            if self.__rpc_codec == "xml":
                return self.__request_xml(methodname, params)
            if self.__rpc_codec == "json":
                return self.__request_json(methodname, params)
        except xmlrpclib.Fault as exc:
            if exc.faultString == "Could not find info-hash.":
                raise HashNotFound(  # pylint: disable=raise-missing-from
                    exc.faultString
                )
            raise exc
        raise ValueError(f"Invalid RPC protocol '{self.__rpc_codec}'")

    def __repr__(self):
        return f"<{self.__class__.__name__} via {self.__rpc_codec} for {self.__url}>"

    def __getattr__(self, name):
        # magic method dispatcher
        # Hardcode the most useful alias
        if name == "log":
            name = "print"
        return xmlrpclib._Method(self.__request, name)

    # note: to call a remote object with a non-standard name, use
    # result getattr(server, "strange-python-name")(args)

    def __call__(self, attr):
        """A workaround to get special attributes on the ServerProxy
        without interfering with the magic __getattr__

        """
        if attr == "close":
            return self.__close
        if attr == "transport":
            return self.__transport
        raise AttributeError(f"Attribute {attr} not found")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__close()



# pylint: disable=too-many-nested-blocks
""" Classification.

    Copyright (c) 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""


import logging
import os
import re

from collections import defaultdict

from pyrosimple import config


log = logging.getLogger(__name__)


KIND_AUDIO = {"flac", "mp3", "ogg", "wav", "dts", "ac3", "alac", "wma"}
KIND_VIDEO = {
    "avi",
    "mkv",
    "m4v",
    "vob",
    "mp4",
    "mpg",
    "mpeg",
    "m2ts",
    "ts",
    "ogv",
    "wmv",
}
KIND_IMAGE = {"jpg", "png", "gif", "tif", "bmp", "svg"}
KIND_DOCS = {
    "chm",
    "pdf",
    "cbr",
    "cbz",
    "odt",
    "ods",
    "doc",
    "xls",
    "ppt",
    "epub",
    "mobi",
    "azw3",
    "djvu",
}
KIND_ARCHIVE = {"rar", "zip", "tgz", "bz2", "iso", "bin"}


_i = _k = None
_VIDEO_EXT = "|".join(re.escape("." + _i) for _i in KIND_VIDEO)
_TV_TRAIL = (
    r"(?:[._ ](?P<release_tags>PREAIR|READNFO))?"
    r"(?:[._ ](?P<release>REPACK|PROPER|REAL|REALPROPER|INTERNAL))?"
    r"(?:[._ ](?P<aspect>WS))?"
    r"(?:[._ ](?P<format>HDTV|PDTV|DSR|DVD[59]?|DVDSCR|480p|576p|720p|1080p|1080i|2160p))?"
    r"(?:[._ ](?P<release2>WEB-DL|WEB\.DL|WEBRip))?"
    r"(?:[._ ](?P<format2>HDTV|PDTV|DSR|DVD[59]?|DVDSCR|480p|576p|720p|1080p|1080i|2160p))?"
    r"(?:[._ ](?P<codec>[XH]\.?264|XviD|VTS|ISO|NTSC|PAL))?"
    r"(?:[._ ](?P<sound>MP3|AC3|DD5\.1|L?PCM|AAC 2\.0))?"
    r"(?:[._ ](?P<codec2>[XH]\.?264|XviD|VTS|ISO|NTSC|PAL))?"
    r"(?:[-. ](?P<group>.+?))?(?P<extension>" + _VIDEO_EXT + ")?$"
)
_DEFINITELY_TV = ["hdtv", "pdtv", "dsr"]

TV_PATTERNS = [
    (_k, re.compile(_i, re.I))
    for _k, _i in (
        (
            "Normal TV Episodes",
            r"^(?P<show>.+?)[._ ]S?(?P<season>\d{1,2})[xE](?P<episode>\d{2}(?:-?E\d{2})?)"
            r"(?:[._ ](?P<title>.+?[a-zA-Z]{1,2}.+?))?" + _TV_TRAIL,
        ),
        (
            "Normal TV Episodes (all-numeric season+episode)",
            r"^(?P<show>.+?)[._ ](?P<season>\d)(?P<episode>\d{2})"
            r"(?:[._ ](?P<title>.+?[a-zA-Z]{1,2}.+?))?" + _TV_TRAIL,
        ),
        (
            "Daily Shows",
            r"^(?P<show>.+?)[._ ](?P<date>\d{4}\.\d{2}\.\d{2})"
            r"(?:[._ ](?P<title>.+?[a-zA-Z]{1,2}.+?))?" + _TV_TRAIL,
        ),
        ("Full Seasons", r"^(?P<show>.+?)[._ ]S?(?P<season>\d{1,2})" + _TV_TRAIL),
        (
            "Mini Series",
            r"^(?P<show>.+?)" r"(?:[._ ](?:Part(?P<part>\d+?)|Pilot)){1,2}"
            #         (?P<year>\d{4})| creates false positives for movies!
            r"(?:[._ ](?P<title>.+?[a-z]{1,2}.+?))??" + _TV_TRAIL,
        ),
        (
            "Mini Series (Roman numerals)",
            r"^(?P<show>.+?)"
            r"(?:[._ ]Pa?r?t[._ ](?P<part>[ivxIVX]{1,3}?))"
            r"(?:[._ ](?P<title>.+?[a-z]{1,2}.+?))??" + _TV_TRAIL,
        ),
    )
]

MOVIE_PATTERNS = [
    (_k, re.compile(_i, re.I))
    for _k, _i in (
        (
            "Scene tagged movie",
            r"^(?P<title>.+?)[. ][\[(]?(?P<year>\d{4})[)\]]?"
            r"(?:[._ ](?P<release>UNRATED|REPACK|INTERNAL|MULTI|PROPER|LIMITED|RERiP))*"
            r"(?:[._ ](?P<format>480p|576p|720p|1080p|1080i|2160p))?"
            r"(?:[._ ](?P<source>BDRip|BRRip|HDRip|DVDRip|PAL|NTSC))"
            r"(?:[._ ](?P<sound1>MP3|AC3|AAC|FLAC|DTS(?:-HD)?))?"
            r"(?:[._ ](?P<codec1>xvid|divx|avc|x264|hevc|h265))?"
            r"(?:[._ ](?P<sound2>MP3|AC3|AAC|FLAC|DTS(?:-HD)?))?"
            # r"(?:[._ ](?P<channels>6ch))?"
            r"(?:[-.](?P<group>.+?))?" r"(?P<extension>" + _VIDEO_EXT + ")?$",
        ),
        (
            "Blu-ray movie",
            r"^(?P<title>.+?)[. ][\[(]?(?P<year>\d{4})[)\]]?"
            r"(?:[._ ](?P<release>UNRATED|REPACK|INTERNAL|MULTI|PROPER|LIMITED|RERiP))*"
            r"(?:[._ ](?P<format0>720p|1080p|1080i|2160p))?"
            r"(?:[._ ](?P<source>Blu-ray|BluRay|BD25|BD50))"
            r"(?:[._ ](?P<format>720p|1080p|1080i|2160p))?"
            r"(?:[._ ](?P<codec1>avc|x264|hevc|h265))?"
            r"(?:[._ ](?P<sound>AC3|AAC|FLAC|DTS(?:-HD)?))*"
            r"(?:[._ ](?P<channels>6ch|MA.5.1))?"
            r"(?:[._ ](?P<codec2>avc|x264|hevc|h265))?"
            r"(?:[-.](?P<group>.+?))?"
            r"(?P<extension>" + _VIDEO_EXT + ")?$",
        ),
    )
]

BAD_TITLE_WORDS = {
    "bdrip",
    "brrip",
    "hdrip",
    "dvdrip",
    "ntsc",
    "hdtv",
    "dvd-r",
    "dvdr",
    "dvd5",
    "dvd9",
    "web-dl",
    "blu-ray",
    "bluray",
    "bd25",
    "bd50",
    "480p",
    "576p",
    "720p",
    "1080p",
    "2160p",
    "mp3",
    "ac3",
    "dts",
}

del _k, _i


def get_filetypes(filelist, path=None, size=os.path.getsize):
    """Get a sorted list of file types and their weight in percent
    from an iterable of file names.

    @return: List of weighted file extensions (no '.'), sorted in descending order
    @rtype: list of (weight, filetype)
    """
    path = path or (lambda _: _)

    # Get total size for each file extension
    histo = defaultdict(int)
    for entry in filelist:
        ext = os.path.splitext(path(entry))[1].lstrip(".").lower()
        if ext and ext[0] == "r" and ext[1:].isdigit():
            ext = "rar"
        elif ext == "jpeg":
            ext = "jpg"
        elif ext == "mpeg":
            ext = "mpg"
        histo[ext] += size(entry)

    # Normalize values to integer percent
    total = sum(histo.values())
    if total:
        for ext, val in histo.items():
            histo[ext] = int(val * 100.0 / total + 0.499)

    return sorted(zip(histo.values(), histo.keys()), reverse=True)


def name_trait(name, add_info=False):
    """Determine content type from name."""
    kind, info = None, {}

    # Anything to check against?
    if name and not name.startswith("VTS_"):
        lower_name = name.lower()
        trait_patterns = (
            ("tv", TV_PATTERNS, "show"),
            ("movie", MOVIE_PATTERNS, "title"),
        )

        # TV check
        if any(i in lower_name for i in _DEFINITELY_TV):
            kind = "tv"
            trait_patterns = trait_patterns[:1]

        # Regex checks
        re_name = ".".join([i.lstrip("[(").rstrip(")]") for i in name.split(" .")])
        for trait, patterns, title_group in trait_patterns:
            matched, patname = None, None

            for patname, pattern in patterns:
                matched = pattern.match(re_name)
                ##print matched, patname, re_name; print "   ", pattern.pattern
                if matched and not any(
                    i in matched.groupdict()[title_group].lower()
                    for i in BAD_TITLE_WORDS
                ):
                    kind, info = trait, matched.groupdict()
                    break

            if matched:
                info["pattern"] = patname

                # Fold auxiliary groups into main one
                for key, val in list(info.items()):
                    if key[-1].isdigit():
                        del info[key]
                        if val:
                            key = re.sub("[0-9]+$", "", key)
                            info[key] = f"{info.get(key) or ''} {val}".strip()
                break

    # Return requested result
    return (kind, info) if add_info else kind


def detect_traits(name=None, alias=None, filetype=None):
    """Build traits list from passed attributes.

    The result is a list of hierarchical classifiers, the top-level
    consisting of "audio", "movie", "tv", "video", "document", etc.
    It can be used as a part of completion paths to build directory
    structures.
    """
    result = []
    if filetype:
        filetype = filetype.lstrip(".")

    # Check for "themed" trackers
    if alias and alias in config.settings.ALIAS_TRAITS:
        result = [config.settings.ALIAS_TRAITS[alias], filetype or "other"]

    # Guess from file extensionn and name
    elif filetype in KIND_AUDIO:
        result = ["audio", filetype]
    elif filetype in KIND_VIDEO:
        result = ["video", filetype]

        contents = name_trait(name)
        if contents:
            result = [contents, filetype]
    elif filetype in KIND_IMAGE:
        result = ["img", filetype]
    elif filetype in KIND_DOCS:
        result = ["docs", filetype]
    elif filetype in KIND_ARCHIVE:
        result = ["misc", filetype]

        contents = name_trait(name)
        if contents:
            result = [contents, filetype]

    return result



""" Metafile Editor.

    Copyright (c) 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import copy
import difflib
import hashlib
import os
import re
import sys
import time
import urllib.parse

from pathlib import Path
from typing import List, Optional, cast

import bencode  # typing: ignore

from pyrosimple import config, error
from pyrosimple.scripts.base import ScriptBase
from pyrosimple.util import metafile


def replace_fields(meta, patterns):
    """Replace patterns in fields."""
    for pattern in patterns:
        try:
            assignment, regex, subst, _ = pattern.split(pattern[-1])
            keypath, _ = metafile.parse_assignment_string(assignment)
            namespace = meta
            for key in keypath[:-1]:
                namespace = namespace[key]
            namespace[keypath[-1]] = re.sub(regex, subst, namespace[keypath[-1]])
        except (KeyError, IndexError, TypeError, ValueError) as exc:
            raise error.UserError(f"Bad substitution '{pattern}' ({exc})!") from exc

    return meta


def diff_metafiles(
    meta1: metafile.Metafile,
    meta2: metafile.Metafile,
    meta1_name="old",
    meta2_name="new",
) -> List[str]:
    """Return a git-style diff between two metafiles"""

    from pyrosimple.util.fmt import (  # pylint: disable=import-outside-toplevel
        BencodeJSONEncoder,
    )

    def encode_meta(meta: metafile.Metafile) -> List[str]:
        """Sanitize and modify meta for better diffing"""
        meta_dict = meta.dict_copy()
        if "info" in meta_dict and "pieces" in meta_dict["info"]:
            meta_dict["info"][
                "pieces"
            ] = f'{len(meta_dict["info"]["pieces"])//20} pieces, md5 hash {hashlib.md5(meta_dict["info"]["pieces"]).hexdigest()}'

        return [
            l + "\n" for l in BencodeJSONEncoder(indent=2).encode(meta_dict).split("\n")
        ]

    meta1_out = encode_meta(meta1)
    meta2_out = encode_meta(meta2)
    return list(difflib.unified_diff(meta1_out, meta2_out, meta1_name, meta2_name))


class MetafileChanger(ScriptBase):
    """Change attributes of a bittorrent metafile."""

    # Keys of rTorrent session data
    RT_RESUME_KEYS = ("libtorrent_resume", "log_callback", "err_callback", "rtorrent")

    ENABLE_PROGRESS = True

    def add_options(self):
        """Add program options."""
        super().add_options()

        self.parser.add_argument("metafile", nargs="+", help="Torrent files to modify")
        self.add_bool_option(
            "-n",
            "--dry-run",
            help="don't write changes to disk, just tell what would happen",
        )
        self.add_bool_option(
            "--diff",
            help="display a diff for each change being made",
        )
        self.add_bool_option(
            "-V",
            "--no-validation",
            "--no-skip",
            help="do not validate metafiles before and after change",
        )
        self.add_value_option(
            "-o",
            "--output-directory",
            "PATH",
            help="optional output directory for the modified metafile(s)",
        )
        self.add_bool_option(
            "-p", "--make-private", help="make torrent private (DHT/PEX disabled)"
        )
        self.add_bool_option(
            "-P", "--make-public", help="make torrent public (DHT/PEX enabled)"
        )
        self.add_value_option(
            "-s",
            "--set",
            "KEY=VAL",
            action="append",
            default=[],
            help="set a specific key to the given value; omit the '=' to delete a key",
        )
        self.add_value_option(
            "-r",
            "--regex",
            "KEY/REGEX/SUBST/",
            action="append",
            default=[],
            help="replace pattern in a specific key by the given substitution",
        )
        self.add_bool_option(
            "-C",
            "--clean",
            help="remove all non-standard data from metafile outside the info dict",
        )
        self.add_bool_option(
            "-A",
            "--clean-all",
            help="remove all non-standard data from metafile including inside the info dict",
        )
        self.add_bool_option(
            "-X",
            "--clean-xseed",
            help="like --clean-all, but keep libtorrent resume information",
        )
        self.add_bool_option(
            "-R",
            "--clean-rtorrent",
            help="remove all rTorrent session data from metafile",
        )
        self.add_value_option(
            "-H",
            "--hashed",
            "--fast-resume",
            "DATAPATH",
            help="add libtorrent fast-resume information (use {} in place of the torrent's name in DATAPATH)",
        )
        self.add_value_option(
            "-c",
            "--check-data",
            "PATH",
            help="check the hash against the data in the given path before making any changes",
        )
        self.add_value_option(
            "-T",
            "--tracker",
            "DOMAIN",
            help="filter given torrents for a tracker domain",
        )
        self.add_value_option(
            "-a",
            "--reannounce",
            "URL",
            help="set a new announce URL, but only if the old announce URL matches the new one",
        )
        self.add_value_option(
            "--reannounce-all",
            "URL",
            help="set a new announce URL on ALL given metafiles",
        )
        self.add_bool_option(
            "--no-cross-seed",
            help="when using --reannounce-all, do not add a non-standard field to the info dict ensuring unique info hashes",
        )
        self.add_value_option(
            "--comment", "TEXT", help="set a new comment (an empty value deletes it)"
        )
        self.add_bool_option("--bump-date", help="set the creation date to right now")
        self.add_bool_option("--no-date", help="remove the 'creation date' field")

    def mainloop(self) -> None:
        """The main loop."""
        self.args = self.options.metafile
        if not self.args:
            self.parser.print_help()
            self.parser.error("No metafiles given, nothing to do!")
            self.parser.exit()

        if self.options.reannounce and self.options.reannounce_all:
            self.parser.error("Conflicting options --reannounce and --reannounce-all!")

        # Set filter criteria for metafiles
        filter_url_prefix: Optional[str] = None
        if self.options.reannounce:
            filter_url_prefix = config.map_announce2alias(self.options.reannounce)
            self.log.info(
                "Filtering for metafiles with announce URL %r...",
                filter_url_prefix,
            )
        elif self.options.tracker:
            filter_url_prefix = config.map_announce2alias(self.options.tracker)
            self.log.info(
                "Filtering for metafiles with tracker %r...",
                filter_url_prefix,
            )
        if self.options.reannounce_all:
            self.options.reannounce = self.options.reannounce_all
        else:
            # When changing the announce URL w/o changing the domain, don't change the info hash!
            self.options.no_cross_seed = True

        # Resolve tracker alias, if URL doesn't look like an URL
        if (
            self.options.reannounce
            and not urllib.parse.urlparse(self.options.reannounce).scheme
        ):
            for conn in config.multi_connection_lookup(self.options.reannounce):
                if urllib.parse.urlparse(conn).scheme:
                    self.options.reannounce = conn
                    break
            else:
                raise error.UserError(
                    "Unknown tracker alias or URL %r!" % (self.options.reannounce)
                )

        # go through given files
        bad = 0
        changed = 0
        for filename in self.options.metafile:
            try:
                # Read and remember current content
                torrent: metafile.Metafile = metafile.Metafile.from_file(Path(filename))
            except (OSError, KeyError, bencode.BencodeDecodeError) as exc:
                # There is no way to recover from this
                self.log.warning(
                    "Skipping bad bencode file %r (%s: %s)",
                    filename,
                    type(exc).__name__,
                    exc,
                )
                bad += 1
                continue

            # Check metafile integrity
            if not self.options.no_validation:
                try:
                    torrent.check_meta()
                except ValueError as exc:
                    self.log.warning(
                        "Metafile %r failed pre-change integrity check: %s",
                        filename,
                        exc,
                    )
                    bad += 1
                    continue

            if self.options.check_data:
                try:
                    from pyrosimple.util.metafile import (  # pylint: disable=import-outside-toplevel
                        PieceFailer,
                    )

                    piece_logger = PieceFailer(torrent, self.log)

                    if self.options.progress == "on":
                        # pylint: disable=import-outside-toplevel
                        from pyrosimple.util.ui import (
                            HashProgressBar,
                            HashProgressBarCounter,
                        )

                        with HashProgressBar() as pb:
                            progress_callback = cast(
                                HashProgressBarCounter, pb()
                            ).progress_callback

                            torrent.hash_check(
                                Path(self.options.check_data),
                                progress_callback=progress_callback,
                                piece_callback=piece_logger.check_piece,
                            )
                    else:
                        torrent.hash_check(
                            Path(self.options.check_data),
                            piece_callback=piece_logger.check_piece,
                        )
                except OSError as exc:
                    self.log.error(
                        "File %s does match data from %s: %s",
                        filename,
                        self.options.check_data,
                        exc,
                    )
                    sys.exit(error.EX_DATAERR)

            # Skip any metafiles that don't meet the pre-conditions
            if (
                filter_url_prefix
                and config.map_announce2alias(torrent["announce"]) != filter_url_prefix
            ):
                self.log.info(
                    "Skipping metafile %r: not tracked by %r!",
                    filename,
                    filter_url_prefix,
                )
                continue

            # Keep resume info safe
            libtorrent_resume = {}
            if "libtorrent_resume" in torrent:
                try:
                    libtorrent_resume["bitfield"] = torrent["libtorrent_resume"][
                        "bitfield"
                    ]
                except KeyError:
                    pass  # nothing to remember

                libtorrent_resume["files"] = copy.deepcopy(
                    torrent["libtorrent_resume"]["files"]
                )

            # Change private flag?
            if self.options.make_private and not torrent["info"].get("private", 0):
                self.log.info("Setting private flag...")
                torrent["info"]["private"] = 1
            if self.options.make_public and torrent["info"].get("private", 0):
                self.log.info("Clearing private flag...")
                del torrent["info"]["private"]

            # Remove non-standard keys?
            if self.options.clean or self.options.clean_all or self.options.clean_xseed:
                torrent.clean_meta(
                    including_info=not self.options.clean,
                )

            # Restore resume info?
            if self.options.clean_xseed:
                if libtorrent_resume:
                    self.log.info("Restoring key 'libtorrent_resume'...")
                    torrent.setdefault("libtorrent_resume", {})
                    torrent["libtorrent_resume"].update(libtorrent_resume)
                else:
                    self.log.warning("No resume information found!")

            # Clean rTorrent data?
            if self.options.clean_rtorrent:
                for key in self.RT_RESUME_KEYS:
                    if key in torrent:
                        self.log.info("Removing key %r...", key)
                        del torrent[key]

            # Change announce URL?
            if self.options.reannounce:
                torrent["announce"] = self.options.reannounce
                if "announce-list" in torrent:
                    del torrent["announce-list"]

                if not self.options.no_cross_seed:
                    # Enforce unique hash per tracker
                    torrent["info"]["x_cross_seed"] = hashlib.md5(
                        self.options.reannounce.encode()
                    ).hexdigest()

            # Change comment or creation date?
            if self.options.comment is not None:
                if self.options.comment:
                    torrent["comment"] = self.options.comment
                elif "comment" in torrent:
                    del torrent["comment"]
            if self.options.bump_date:
                torrent["creation date"] = int(time.time())
            if self.options.no_date and "creation date" in torrent:
                del torrent["creation date"]

            # Add fast-resume data?
            if self.options.hashed:
                datadir: str = self.options.hashed
                if "{}" in datadir and not os.path.exists(datadir):
                    datadir = datadir.replace("{}", torrent["info"]["name"])
                try:
                    torrent.add_fast_resume(Path(datadir))
                except OSError as exc:
                    self.fatal("Error making fast-resume data (%s)", exc)
                    raise

            # Set specific keys?
            torrent.assign_fields(self.options.set)
            replace_fields(torrent, self.options.regex)

            # Validate the new data
            if not self.options.no_validation:
                try:
                    torrent.check_meta()
                except ValueError as exc:
                    self.log.warning(
                        "Metafile %r failed post-change integrity check: %s",
                        filename,
                        exc,
                    )
                    bad += 1
                    continue

            # Write new metafile, if changed
            if self.options.output_directory:
                filename = Path(
                    self.options.output_directory, os.path.basename(filename)
                )
                self.log.info("Will write %r...", filename)

                if not self.options.dry_run:
                    torrent.save(filename)
                    if "libtorrent_resume" in torrent:
                        # Also write clean version
                        filename = Path(
                            str(filename).replace(".torrent", "-no-resume.torrent")
                        )
                        del torrent["libtorrent_resume"]
                        self.log.info(
                            "Writing file without libtorrent resume '%s'...", filename
                        )
                        torrent.save(filename)
            else:
                current_torrent = metafile.Metafile.from_file(filename)
                diff = diff_metafiles(current_torrent, torrent, filename)
                if self.options.diff:
                    sys.stdout.writelines(diff)
                if not diff:
                    self.log.info("No change to file %r", filename)
                    continue
                self.log.info("Changing %r...", filename)

                if not self.options.dry_run:
                    try:
                        self.log.debug("Writing file '%s'...", filename)
                        torrent.save(Path(filename))
                    except OSError as exc:
                        raise error.LoggableError(
                            "Can't write to file %s (%s)" % (filename, exc)
                        ) from exc

            changed += 1

        # Print summary
        if changed:
            self.log.info(
                "%s %d metafile(s).",
                "Would've changed" if self.options.dry_run else "Changed",
                changed,
            )
        if bad:
            self.log.warning("Skipped %d bad metafile(s)!", bad)


def run():  # pragma: no cover
    """The entry point."""
    MetafileChanger().run()


if __name__ == "__main__":
    run()



# pylint: disable=attribute-defined-outside-init
""" rTorrent queue manager & daemon.

    Copyright (c) 2012 The PyroScope Project <pyroscope.project@gmail.com>
"""

import errno
import logging
import os
import signal
import sys
import time

from pathlib import Path
from typing import Dict

from apscheduler.schedulers.background import BackgroundScheduler
from box.box import Box
from daemon import DaemonContext
from daemon.pidfile import TimeoutPIDLockFile
from lockfile.pidlockfile import AlreadyLocked, LockFailed

from pyrosimple import config, error
from pyrosimple.scripts.base import ScriptBaseWithConfig
from pyrosimple.util import pymagic


def pid_exists(pid):
    """Check whether pid exists in the current process table."""
    if pid == 0:
        return True
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True
        raise
    return True


class RtorrentQueueManager(ScriptBaseWithConfig):
    """
    rTorrent queue manager & daemon.
    """

    POLL_TIMEOUT = 1.0

    RUNTIME_DIR = os.getenv("XDG_RUNTIME_DIR") or "~/.pyrosimple/run/"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.classes = {}
        self.jobs: Dict = {}
        self.running_config = {}

    def add_options(self):
        """Add program options."""
        super().add_options()

        # basic options
        self.add_bool_option(
            "-n",
            "--dry-run",
            help="advise jobs not to do any real work, just tell what would happen",
        )
        self.add_bool_option(
            "--no-fork",
            "--fg",
            help="Don't fork into background (stay in foreground and log to console)",
        )
        self.add_value_option(
            "--run-once", "JOB", help="run the specified job once in the foreground"
        )
        self.add_bool_option("--stop", help="stop running daemon")
        self.add_bool_option(
            "--restart",
            help="stop any existing daemon, then start the process in the backgrounds",
        )
        self.add_bool_option("-?", "--status", help="Check daemon status")
        self.add_value_option(
            "--pid-file",
            "PATH",
            help="file holding the process ID of the daemon, when running in background",
            type=Path,
        )
        self.add_bool_option(
            "--adopt-stale-pid-file",
            help="if the pid file exists but appears to be stale, adopt it",
        )
        self.add_value_option(
            "--log-file",
            "PATH",
            type=Path,
            help="file for logging stderr/stdout of the forked process (only used when running the daemon in the background)",
        )

    def parse_schedule(self, schedule):
        """Parse a job schedule."""
        result = {}

        for param in schedule.split():
            param = param.strip()
            try:
                key, val = param.split("=", 1)
                if key == "jitter":
                    val = int(val)
            except (TypeError, ValueError) as exc:
                raise error.ConfigurationError(
                    f"Bad param {param!r} in job schedule {schedule!r}"
                ) from exc
            result[key] = val

        return result

    def validate_config(self):
        """Handle and check configuration."""

        for name, params in config.settings.TORQUE.items():
            # Skip non-dictionary keys
            if not isinstance(params, Box) or name == "_settings":
                continue
            for key in ("handler", "schedule"):
                if key not in params:
                    raise error.ConfigurationError(
                        f"Job {name!r} is missing the required {key!r} parameter"
                    )
            self.jobs[name] = dict(params)
            if self.options.dry_run:
                self.jobs[name]["dry_run"] = True
            if params.get("active", True):
                self.jobs[name]["__handler"] = pymagic.import_name(params.handler)
            self.jobs[name]["schedule"] = self.parse_schedule(params.get("schedule"))

    def add_jobs(self):
        """Add configured jobs."""
        for name, params in self.jobs.items():
            if params.get("active", True):
                params.setdefault("__job_name", name)
                # Keep track of the instantiated class for cleanup later
                self.classes[name] = params["__handler"](params)
                self.sched.add_job(
                    self.classes[name].run,
                    name=name,
                    id=name,
                    trigger="cron",
                    **params["schedule"],
                )

    def unload_jobs(self):
        """Allows jobs classes to clean up any global resources if the
        cleanup() method exists.

        This should be called only once the jobs have finished
        running, so that a successive run doesn't re-create the
        resources.
        """
        for _, cls in self.classes.items():
            if hasattr(cls, "cleanup") and callable(cls.cleanup):
                cls.cleanup()

    def reload_jobs(self):
        """Reload the configured jobs gracefully."""
        try:
            config.load_settings()
            if self.running_config != dict(config.settings.TORQUE):
                self.log.info("Config change detected, reloading jobs")
                self.validate_config()
                self.sched.pause()
                self.sched.remove_all_jobs()
                self.unload_jobs()
                self.sched.resume()
                self.add_jobs()
                self.running_config = dict(config.settings.TORQUE)
        except Exception as exc:  # pylint: disable=broad-except
            self.log.error("Error while reloading config: %s", exc)
        else:
            self.sched.resume()

    def run_forever(self):
        """Run configured jobs until termination request."""
        self.running_config = dict(config.settings.TORQUE)
        reload_config = config.settings.TORQUE.get(
            "autoreload", False
        ) or config.settings.TORQUE._settings.get("autoreload", False)
        while True:
            try:
                time.sleep(self.POLL_TIMEOUT)
                if reload_config:
                    self.reload_jobs()
            except KeyboardInterrupt as exc:
                self.log.info("Termination request received (%r)", exc)
                self.sched.shutdown()
                self.unload_jobs()
                break
            except SystemExit as exc:
                if isinstance(exc.code, int):
                    self.return_code = exc.code
                else:
                    self.return_code = 0
                self.log.info("System exit (RC=%r)", self.return_code)
                break

    def mainloop(self):
        """The main loop."""
        try:
            self.validate_config()
        except error.ConfigurationError as exc:
            self.fatal(exc)

        # Defaults for process control paths
        pid_file = TimeoutPIDLockFile(
            Path(
                self.options.pid_file
                or config.settings.TORQUE._settings.get(
                    "pid_file", Path(self.RUNTIME_DIR, "pyrotorque.pid")
                )
            )
        )
        log_file = self.options.log_file or config.settings.TORQUE._settings.get(
            "log_file", None
        )
        log_level = (
            config.settings.TORQUE._settings.get("log_level", None)
            or self.options.log_level
        )

        # Process control
        if pid_file.is_locked():
            running, pid = True, pid_file.read_pid()
        else:
            running, pid = False, 0
        if self.options.status or self.options.stop or self.options.restart:
            if self.options.status:
                if running:
                    if pid_exists(pid):
                        self.log.info("Pyrotorque is running (PID %d).", pid)
                        sys.exit(0)
                    else:
                        self.log.error(
                            "PID file exist, but process %d appears stale", pid
                        )
                        sys.exit(1)
                else:
                    self.log.error("No pyrotorque process found.")
                    sys.exit(1)

            if self.options.stop or self.options.restart:
                if running:
                    os.kill(pid, signal.SIGTERM)
                    self.log.debug("Process %d sent SIGTERM.", pid)

                    # Wait for termination (max. 10 secs)
                    for _ in range(100):
                        if not pid_file.is_locked():
                            running = False
                            break
                        time.sleep(0.1)

                    self.log.info("Process %d stopped.", pid)
                elif pid:
                    self.log.info("Process %d NOT running anymore.", pid)
                else:
                    self.log.info("No pid file '%s'", (pid_file.path or "<N/A>"))
            else:
                self.log.info(
                    "Process %d %s running.", pid, "UP and" if running else "NOT"
                )

            if self.options.stop:
                self.return_code = error.EX_OK if running else error.EX_UNAVAILABLE
                return

        # Check if we only need to run once
        if self.options.run_once:
            try:
                params = self.jobs[self.options.run_once]
            except KeyError:
                self.log.critical(
                    "Job %r not found in configuration file!", self.options.run_once
                )
                sys.exit(error.EX_SOFTWARE)
            if self.options.dry_run:
                params["dry_run"] = True
            # Make a copy here to prevent the original handler from
            # getting overwritten
            params["__handler_copy"] = params.get("__handler")(params)
            params["__handler_copy"].run()
            sys.exit(0)

        dcontext = DaemonContext(
            detach_process=False,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        # Set up for forking into the background, if not disabled via
        # the flag
        if not self.options.no_fork:
            dcontext.stdin = None
            if log_file is not None:
                log_file_handle = open(  # pylint: disable=consider-using-with
                    log_file, "w", encoding="utf-8"
                )
            else:
                log_file_handle = None
            dcontext.stderr = log_file_handle
            dcontext.stdout = log_file_handle
            dcontext.pidfile = pid_file
            if running and self.options.adopt_stale_pid_file and not pid_exists(pid):
                self.log.debug("Removing stale PID file")
                os.unlink(pid_file.path)
                with pid_file:
                    pass
            # Ensure we can lock the pid_file ahead of time
            try:
                with pid_file:
                    pass
            except (AlreadyLocked, LockFailed) as exc:
                self.log.error("Cannot lock pidfile: %s", exc)
                sys.exit(1)
            dcontext.detach_process = True
            self.log.info("Writing pid to %s and detaching process...", pid_file.path)
            self.log.info("Logging stderr/stdout to %r", log_file or "/dev/null")

        # Change logging format to something more daemon-like
        logging.basicConfig(
            force=True,
            format=config.settings.TORQUE._settings.get(
                "log_format", "%(asctime)s %(levelname)5s %(name)s: %(message)s"
            ),
        )
        self.log.setLevel(log_level)

        with dcontext:
            # Set up services
            self.sched = BackgroundScheduler()

            # Run services
            self.sched.start()
            try:
                self.add_jobs()
                self.run_forever()
            finally:
                if self.sched.running:
                    self.log.info("Shutting down scheduler...")
                    self.sched.shutdown()


def run():  # pragma: no cover
    """The entry point."""
    RtorrentQueueManager().run()


if __name__ == "__main__":
    run()



""" Administration Tool.

    Copyright (c) 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""

import importlib.resources
import re

from datetime import datetime
from pathlib import Path
from xmlrpc import client as xmlrpclib

import tomli_w

import pyrosimple

from pyrosimple import config
from pyrosimple.scripts.base import ScriptBase
from pyrosimple.util import matching


class AdminTool(ScriptBase):
    """Support for administrative tasks."""

    # TODO: config create, dump, set, get
    # TODO: backup session/config

    def add_options(self):
        super().add_options()
        self.parser.add_argument("-U", "--url", help="URL to rtorrent instance")
        self.parser.set_defaults(func=None)
        subparsers = self.parser.add_subparsers()
        config_parser = subparsers.add_parser("config", help="Validate configuration")
        config_parser.set_defaults(func=self.config)
        config_parser.add_argument(
            "--check", help="Check config for any issues", action="store_true"
        )
        config_parser.add_argument(
            "--dump-rc", help="Print out the full rTorrent config", action="store_true"
        )
        config_parser.add_argument(
            "--create-config",
            help="Create config.toml in the default location if it does not exist",
            action="store_true",
        )
        config_parser.add_argument(
            "--create-rtorrent-rc",
            help="Create a rtorrent.rc in the default location if it does not exist",
            action="store_true",
        )
        backfill_parser = subparsers.add_parser(
            "backfill", help="Backfill missing custom fields from available data"
        )
        backfill_parser.set_defaults(func=self.backfill)
        backfill_parser.add_argument(
            "--dry-run",
            help="Print changes instead of applying them",
            action="store_true",
        )

    def dump_rc(self):
        """Print a representative .rtorrent.rc as gleaned from a running instance.

        This is neat but somewhat brittle, and behaves differently between XMLRPC and JSON-RPC.
        """

        proxy = pyrosimple.connect().open()
        methods = proxy.system.listMethods()
        # XXX This is a heuristic and might break in newer rTorrent versions!
        builtins = set(methods[: methods.index("view.sort_new") + 1])
        methods = set(methods)
        plain_re = re.compile(r"^[a-zA-Z0-9_.]+$")
        rc_continuation_threshold = 50

        def is_method(name):
            "Helper"
            prefixes = (
                "d.",
                "f.",
                "p.",
                "t.",
                "choke_group.",
                "session.",
                "system.",
                "throttle.",
                "trackers.",
                "ui.",
                "view.",
            )

            if name.endswith("="):
                name = name[:-1]
            return plain_re.match(name) and (
                name in methods or any(name.startswith(x) for x in prefixes)
            )

        def rc_quoted(text, in_brace=False):
            "Helper"
            if isinstance(text, list):
                wrap_fmt = "{%s}"
                try:
                    method_name = text[0] + ""
                except (TypeError, IndexError):
                    pass
                else:
                    if is_method(method_name):
                        wrap_fmt = "(%s)" if in_brace else "((%s))"
                        if (
                            ".set" not in method_name
                            and len(text) == 2
                            and text[1] == 0
                        ):
                            text = text[:1]
                text = wrap_fmt % ", ".join(
                    [rc_quoted(x, in_brace=(wrap_fmt[0] == "{")) for x in text]
                )
                return text.replace("))))", ")) ))")
            if isinstance(text, int):
                return f"(value, {text:d})"
            if plain_re.match(text) or is_method(text):
                return text
            return '"{}"'.format(text.replace("\\", "\\\\").replace('"', '\\"'))

        group = None
        for name in sorted(methods):
            try:
                value = proxy.method.get("", name)
                const = bool(proxy.method.const("", name))
            except xmlrpclib.Fault as exc:
                if "Key not found" in exc.faultString:
                    continue
                raise
            group, old_group = name.split(".", 1)[0], group
            if group == "event":
                group = name
            if group != old_group:
                print("")

            definition = None
            objtype = type(value)
            if objtype is list:
                value = [rc_quoted(x) for x in value]
                wrap_fmt = "((%s))" if value and is_method(value[0]) else "{%s}"
                definition = wrap_fmt % ", ".join(value)
            elif objtype is dict:
                print(f"method.insert = {name}, multi|rlookup|static")
                for key, val in sorted(value.items()):
                    val = rc_quoted(val)
                    if len(val) > rc_continuation_threshold:
                        val = "\\\n    " + val
                    print(f'method.set_key = {name}, "{key}", {val}')
            elif objtype is str:
                definition = rc_quoted(value)
            elif objtype is int:
                definition = f"{value:d}"
            else:
                self.log.error(
                    "Cannot handle %r definition of method %s", objtype, name
                )
                continue

            if definition:
                if name in builtins:
                    print(f"{name}.set = {definition}")
                else:
                    rctype = {str: "string", int: "value"}.get(objtype, "simple")
                    if const:
                        rctype += "|const"
                        const = None
                    if len(definition) > rc_continuation_threshold:
                        definition = "\\\n    " + definition
                    definition = definition.replace(" ;     ", " ;\\\n     ").replace(
                        ",    ", ",\\\n    "
                    )
                    print(f"method.insert = {name}, {rctype}, {definition}")
            if const:
                print(f"method.const.enable = {name}")

    def backfill(self):
        """Backfill missing any missing metadata from available sources.
        Safe to run multiple times.
        """
        # pylint: disable=broad-except
        if self.options.url:
            config.settings["SCGI_URL"] = config.lookup_connection_alias(
                self.options.url
            )
        config.settings["FAST_QUERY"] = 0
        engine = pyrosimple.connect()
        engine.open()
        for i in engine.view("main", matching.create_matcher("loaded=0 metafile=/.+/")):
            try:
                mtime = int(Path(i.metafile).stat().st_mtime)
                if self.options.dry_run:
                    dt = datetime.fromtimestamp(mtime)
                    self.log.info(
                        "Would set %s tm_loaded to %s from metafile %s",
                        i.hash,
                        dt,
                        i.metafile,
                    )
                else:
                    i.rpc_call("d.custom.set", ["tm_loaded", str(mtime)])
                    i.flush()
            except Exception as e:
                self.log.error("Could not set tm_loaded for %s: %s", i.hash, e)
        for i in engine.view("main", matching.create_matcher("loaded=0 path=/.+/")):
            try:
                mtime = int(Path(i.path).stat().st_mtime)
                if self.options.dry_run:
                    dt = datetime.fromtimestamp(mtime)
                    self.log.info(
                        "Would set %s tm_loaded to %s from path %s", i.hash, dt, i.path
                    )
                else:
                    i.rpc_call("d.custom.set", ["tm_loaded", str(mtime)])
                    i.flush()
            except Exception as e:
                self.log.error("Could not set tm_loaded for %s: %s", i.hash, e)
        for i in engine.view(
            "main", matching.create_matcher("completed=0 is_complete=yes path=/.+/")
        ):
            try:
                mtime = int(Path(i.path).stat().st_mtime)
                if self.options.dry_run:
                    dt = datetime.fromtimestamp(mtime)
                    self.log.info(
                        "Would set %s tm_completed to %s from path %s",
                        i.hash,
                        dt,
                        i.path,
                    )
                else:
                    i.rpc_call("d.custom.set", ["tm_completed", str(mtime)])
                    i.flush()
            except Exception as e:
                self.log.error("Could not set tm_loaded for %s: %s", i.hash, e)

    def create_config(self):
        """Create a configuration file"""
        config_path = Path(config.SETTINGS_FILE)
        if config_path.exists():
            self.log.info(
                "Pyrosimple config path %s already exists, not overwriting", config_path
            )
        else:
            self.log.info("Creating pyrosimple config file '%s'", config_path)
            config_path.parent.mkdir(parents=True, exist_ok=True)
            with config_path.open("wb") as fh:
                tomli_w.dump(pyrosimple.config.settings.to_dict(), fh)

    def create_rtorrent_rc(self):
        """Create a rtorrent.rc file"""
        rtorrent_rc_path = Path(pyrosimple.config.settings.RTORRENT_RC).expanduser()
        if rtorrent_rc_path.exists():
            self.log.info(
                "rtorrent.rc path %s already exists, not overwriting", rtorrent_rc_path
            )
        else:
            self.log.info("Creating rtorrent.rc at '%s'", rtorrent_rc_path)
            home = str(Path("~").expanduser())
            rtorrent_rc_path.parent.mkdir(parents=True, exist_ok=True)
            with rtorrent_rc_path.open("w", encoding="utf-8") as fh:
                fh.write(
                    importlib.resources.open_text("pyrosimple.data", "full-example.rc")
                    .read()
                    .replace("/home/USERNAME", home)
                )

    def config(self):
        """Handle the config subcommand"""
        if self.options.dump_rc:
            self.dump_rc()
        if self.options.create_config:
            self.create_config()
        if self.options.create_rtorrent_rc:
            self.create_rtorrent_rc()
        if self.options.check:
            if self.options.url:
                config.settings["SCGI_URL"] = config.lookup_connection_alias(
                    self.options.url
                )
            try:
                config.autoload_scgi_url()
            except Exception:
                self.log.error("Error loading SCGI URL:")
                raise
            self.log.debug("Loaded SCGI URL successfully")
            try:
                pyrosimple.connect().open()
            except ConnectionRefusedError:
                self.log.error(
                    "SCGI URL '%s' found, but rTorrent may not be running!",
                    config.autoload_scgi_url(),
                )
                raise
            self.log.info("Connected to rTorrent successfully")
            self.config_check_timestamps()

    def config_check_timestamps(self):
        """Confirm usual methods for timestamps exist"""
        proxy = pyrosimple.connect().open()
        methods = proxy.system.listMethods()
        event_field_map = {
            "event.download.resumed": "started",
            "event.download.finished": "completed",
            "event.download.inserted_new": "loaded",
        }
        print_timestamp_help = False
        for event, field in event_field_map.items():
            if "!time_stamp" not in proxy.method.get("", event):
                self.log.warning(
                    "Method '!time_stamp' not found in '%s', field '%s' may not function correctly",
                    event,
                    field,
                )
                print_timestamp_help = True
        if "pyro.last_xfer.min_rate" not in methods:
            self.log.warning(
                "Method 'pyro.last_xfer.min_rate' not found, field 'last_xfer' may not function correctly"
            )
            print_timestamp_help = True
        if "d.timestamp.last_active" not in methods:
            self.log.warning(
                "Method 'd.timestamp.last_active' not found, field 'last_active' may not function correctly"
            )
            print_timestamp_help = True
        if print_timestamp_help:
            self.log.warning(
                "To configure timestamp fields, see https://kannibalox.github.io/pyrosimple/rtorrent-config/#timestamps"
            )

    def mainloop(self):
        if self.options.func is None:
            self.parser.print_help()
            return
        self.options.func()


def run():  # pragma: no cover
    """The entry point."""
    AdminTool().run()


if __name__ == "__main__":
    run()



""" Configuration.

    For details, see https://pyrosimple.readthedocs.io/en/latest/setup.html

    Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>
"""

import functools
import logging
import os
import re
import shlex
import urllib

from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple, Union


try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

from box.box import Box

from pyrosimple import error


ENVVAR = "PYRO_CONF"
ENVVAR_PREFIX = "PYRO"
SETTINGS_FILE = Path("~/.config/pyrosimple/config.toml").expanduser()

DEFAULT_SETTINGS = Box(
    {
        "SCGI_URL": "",
        "RTORRENT_RC": "~/.rtorrent.rc",
        "CONFIG_PY": "~/.config/pyrosimple/config.py",
        "CONFIG_PY_LOADED": False,
        "SORT_FIELDS": "name,hash",
        "FAST_QUERY": 0,
        "ITEM_CACHE_EXPIRATION": 5.0,
        "SAFETY_CHECKS_ENABLED": True,
        "MKTOR_IGNORE": [
            "core",
            "CVS",
            ".*",
            "*~",
            "*.swp",
            "*.tmp",
            "*.bak",
            "[Tt]humbs.db",
            "[Dd]esktop.ini",
            "ehthumbs_vista.db",
        ],
        # TOML sections
        "ALIASES": {},
        "ALIAS_TRAITS": {},
        "CONNECTIONS": {},
        "TORQUE": Box({"_settings": Box()}),
        # Allow individual overrides in FORMATS section
        "FORMATS": {
            "default": '{{d.name}} \t[{{d.alias}}]\n  {{d.is_private|fmt("is_private")}} {{d.is_open|fmt("is_open")}} {{d.is_active|fmt("is_active")}} P{{d.prio|int}} {%if d.is_complete %}     done{%else%}{{"%8.2f"|format(d.done)}}%{%endif%}\t{{d.size|sz}} U:{{d.up|sz}}/s  D:{{d.down|sz}}/s T:{{d.throttle|fmt("throttle")}}',
            "short": '{%set ESC = "\x1B" %}{%if d.down > 0%}{{ESC+"[1m"}}{%endif%}{%if d.is_open%}O{%else%} {%endif%}{%if  d.is_active%}A{%else%} {%endif%}{%if not d.is_complete%}{{ESC+"[36m"}}{{ "{:>3}".format(d.done | round | int) }}{{ESC+"[0m"}}{%else%}  D{%endif%} {{"{:>10}".format(d.size | filesizeformat(True))}} {%if d.message%}{{ESC+"[31m"}}{%endif%} {{d.alias.rjust(3)}}{{ESC+"[0m"}} {%if d.down > 0%}{{ESC+"[1m"}}{%endif%}{{d.name}}{{ESC+"[0m"}}',
            "filelist": "{% for f in d.files %}{{d.realpath}}{% if d.is_multi_file %}/{{f.path}}{% endif %}{% if loop.index != loop.length %}\n{% endif %}{% endfor %}",
            "action": "{{now()|iso}} {{action}}\t {{d.name}} [{{d.alias}}]",
        },
    }
)


def load_settings() -> Box:
    """Load settings from (in order of precedence): the defaults, a
    TOML config file, environment variables
    """
    settings_box: Box = DEFAULT_SETTINGS.copy()
    settings_file = Path(os.getenv(ENVVAR, str(SETTINGS_FILE))).expanduser()
    if settings_file.exists():
        settings_file_box = Box(
            {
                k.upper(): v
                for k, v in tomllib.loads(settings_file.read_text("utf-8")).items()
            }
        )
        settings_box.merge_update(settings_file_box)
    env_settings = {}
    for k, v in os.environ.items():
        if k.startswith(f"{ENVVAR_PREFIX}_"):
            key = k[len(ENVVAR_PREFIX) + 1 :]
            env_settings[key] = v
    settings_box.merge_update(Box(env_settings))
    return settings_box


settings: Box = load_settings()


class RCLexer(shlex.shlex):
    """Helper to split argument lists."""

    def __init__(self, text: str):
        super().__init__(text)
        self.whitespace += ","
        self.whitespace_split = True
        self.commenters = ""


def expand_rc(rcfile: Path) -> List[Tuple[str, str]]:
    """Return key/val pairs for each line in the rc file, with some
    very naive string replacement for cat= calls"""
    log = logging.getLogger(__name__)
    data: List[Tuple[str, str]] = []
    replacements: Dict[str, str] = {}
    with rcfile.open("r", encoding="utf-8") as handle:
        continued = False
        for line in handle.readlines():
            # Skip comments, continuations, and empty lines
            line = line.strip()
            continued, was_continued = line.endswith("\\"), continued
            if not line or was_continued or line.startswith("#"):
                continue

            # Be lenient about errors, after all it's not our own config file
            try:
                key, val = line.split("=", 1)
            except ValueError:
                log.debug("Ignored invalid line %r in %r!", line, rcfile)
                continue
            key, val = key.strip(), val.strip()
            priv_string_match = re.match(
                r"([.\w]+),\s*private\|const\|string,\s*\(cat,(.*)\)", val
            )
            if priv_string_match:
                str_result = ""
                for arg in RCLexer(priv_string_match.group(2)):
                    if arg.startswith("("):
                        str_result += replacements.get(arg[1:-1], "")
                    if arg.startswith('"'):
                        str_result += arg[1:-1]
                val = str_result
                replacements[priv_string_match.group(1)] = str_result
            if key in [
                "network.scgi.open_local",
                "scgi_local",
                "scgi_port",
                "network.scgi.open_port",
            ]:
                simple_cat_match = re.match(r"\(cat,(.*)\)", val)
                if simple_cat_match:
                    str_result = ""
                    for arg in RCLexer(simple_cat_match.group(1)):
                        if arg.startswith("("):
                            str_result += replacements.get(arg[1:-1], "")
                        if arg.startswith('"'):
                            str_result += arg[1:-1]
                    val = str_result
            data.append((key, val))
        return data


def scgi_url_from_rtorrentrc(rcfile: Union[str, Path]) -> Optional[str]:
    """Parse a rtorrent.rc file and get the SCGI connection information"""
    log = logging.getLogger(__name__)
    log.debug("Loading rtorrent config from '%s'", rcfile)
    scgi_local: str = ""
    scgi_port: str = ""
    rcfile = Path(rcfile)
    for key, val in expand_rc(rcfile):
        # Copy values we're interested in
        if key in ["network.scgi.open_port", "scgi_port"]:
            log.debug("rtorrent.rc: %s = %s", key, val)
            scgi_port = val
        if key in ["network.scgi.open_local", "scgi_local"]:
            log.debug("rtorrent.rc: %s = %s", key, val)
            scgi_local = val

    # Validate fields
    if scgi_local and not scgi_local.startswith("scgi+unix://"):
        scgi_local = "scgi+unix://" + str(Path(scgi_local).expanduser())
    if scgi_port and not scgi_port.startswith("scgi://"):
        scgi_port = "scgi://" + scgi_port

    return scgi_local or scgi_port
    # Prefer UNIX domain sockets over TCP sockets


def autoload_scgi_url() -> str:
    """Load and return the SCGI URL, auto-resolving it if necessary."""
    if settings.SCGI_URL:
        return str(settings.SCGI_URL)
    # Get and check config file name
    rcfile = Path(settings.RTORRENT_RC).expanduser()
    if not rcfile.exists():
        raise error.UserError(f"rTorrent RC file '{rcfile}' doesn't exist!")
    scgi_url = scgi_url_from_rtorrentrc(rcfile)

    settings["SCGI_URL"] = scgi_url

    return str(settings.SCGI_URL)


def lookup_announce_alias(name: str):
    """Get canonical alias name and announce URL list for the given
    alias."""
    for alias, urls in settings["ALIASES"].items():
        if alias.lower() == name.lower():
            return alias, urls

    raise KeyError(f"Unknown alias {name}")


def lookup_announce_url(name: str) -> Tuple[str, List[str]]:
    """Get canonical alias name and announce URL list for the given
    alias.

    Unlike lookup_announce_alias, only valid URLs are returned
    """
    for alias, urls in settings["ALIASES"].items():
        if alias.lower() == name.lower():
            result = []
            for url in urls:
                if urllib.parse.urlparse(url).scheme:
                    result.append(url)
            return alias, result

    raise KeyError(f"Unknown alias {name}")


@functools.lru_cache(maxsize=None)
def map_announce2alias(url: str) -> str:
    """Get tracker alias for announce URL, and if none is defined, the
    2nd level domain."""
    url = url.strip()
    if url in settings["ALIASES"].items():
        return url
    # Try to find an exact alias URL match and return its label
    for alias, urls in settings["ALIASES"].items():
        if any(i.strip() == url for i in urls):
            return str(alias)

    # Try to find an alias URL prefix and return its label
    parts = urllib.parse.urlparse(url)
    server = urllib.parse.urlunparse(
        (parts.scheme, parts.netloc, "/", None, None, None)
    )

    for alias, urls in settings["ALIASES"].items():
        if any(i.strip().startswith(server) for i in urls):
            return str(alias)

    # Return 2nd level domain name if no alias found
    try:
        # Try to find based on domain
        domain = ".".join(parts.netloc.split(":")[0].split(".")[-2:])
        for alias, urls in settings["ALIASES"].items():
            if any(i.strip() == domain for i in urls):
                return str(alias)
        return domain
    except IndexError:
        return parts.netloc


def load_custom_py():
    """Load custom python configuration.

    This only gets called manually to prevent some weird code
    injection if pyrosimple is ever used in a library.
    """
    log = logging.getLogger(__name__)
    if not settings.CONFIG_PY:
        log.debug("Custom code loading is disabled")
        return
    if settings.CONFIG_PY_LOADED:
        log.debug("Custom code has already been loaded")
    config_file = Path(settings.CONFIG_PY).expanduser()
    if config_file.exists():
        log.debug("Loading '%s'...", config_file)
        with open(config_file, "rb") as handle:
            # pylint: disable=exec-used
            exec(handle.read())
        settings.CONFIG_PY_LOADED = True
    else:
        log.debug("Configuration file '%s' not found.", config_file)


def lookup_connection_alias(url: str) -> str:
    """Convert a connection alias to the actual URL (if set in the
    config"""
    if url in settings["CONNECTIONS"]:
        return str(settings["CONNECTIONS"][url])
    return url


def multi_connection_lookup(url: str) -> Iterator[str]:
    """Return a list of urls.

    This is separate from lookup_connection_alias due to scripts
    needing to be written specifically to handle this.
    """
    val = settings["CONNECTIONS"].get(url, [url])
    if isinstance(val, list):
        for v in val:
            yield lookup_connection_alias(v)
    else:
        yield lookup_connection_alias(val)



"""Jobs for metric reporting"""
from collections import Counter
from threading import Thread
from time import sleep, time
from typing import Any, Dict, List, Optional

from prometheus_client import REGISTRY, exposition
from prometheus_client.core import Counter as PromCounter
from prometheus_client.core import Gauge, Info

from pyrosimple import error
from pyrosimple.job.base import BaseJob
from pyrosimple.util import fmt, rpc


class EngineStats(BaseJob):
    """Simple rTorrent connection statistics logger."""

    def run(self):
        """Statistics logger job callback."""
        try:
            self.engine.open()
            self.log.info(
                "Stats for %s - up %s, %s torrents",
                self.engine.engine_id,
                fmt.human_duration(
                    self.engine.rpc.system.time() - self.engine.startup, 0, 2, True
                ).strip(),
                self.engine.rpc.view.size(rpc.NOHASH, "default"),
            )
        except (error.LoggableError, *rpc.ERRORS) as exc:
            self.log.warning(str(exc))


class RtorrentExporter(BaseJob):
    """Expose rTorrent and host statistics for scraping by a
    Prometheus instance."""

    def add_metric(self, name, cls, *args, **kwargs):
        """Only add the metric if it doesn't exist"""
        if name not in self.metrics:
            self.metrics[name] = cls(*args, **kwargs)
        return self.metrics[name]

    def __init__(self, config: Optional[Dict] = None):
        """Set up RtorrentExporter."""
        super().__init__(config)
        self.prefix = self.config.get("prefix", "rtorrent_")
        self.proxy = self.engine.open()

        self.active_jobs = set()
        jobs = ["item", "tracker", "system"]
        for j in self.config.get("jobs", "system").split(","):
            j = j.strip()
            if j not in jobs:
                self.log.warning("Job %s not found, skipping", j)
            else:
                self.active_jobs.add(j)
        self.metrics: Dict[str, Any] = {}

        # Create "constant" info metric
        self.info_methods = ["system.client_version", "system.library_version"]
        system_info = self.add_metric(
            "info", Info, self.prefix.rstrip("_"), "rTorrent platform information"
        )
        info_call = [
            {"methodName": method, "params": []} for method in self.info_methods
        ]
        info_results = [r[0] for r in self.proxy.system.multicall(info_call)]
        system_info.info(
            dict(zip([m.replace(".", "_") for m in self.info_methods], info_results))
        )

        # Set up some class variables that may not be used (depending
        # on configured jobs) for typing purposes
        self.system_stats: List[str] = []
        self.views: List[str] = []
        self.item_stats = sorted(
            set(self.config.get("item-stats", ["d.down.total", "d.up.total"]))
        )
        self.item_labels = sorted(
            set(self.config.get("item-labels", ["d.hash", "d.name"]))
        )
        self.add_metric(
            "scrape",
            PromCounter,
            self.prefix + "scrape",
            "Amount of scrapes against rTorrent",
        )
        self.add_metric(
            "last_scrape",
            Gauge,
            self.prefix + "last_scrape",
            "Timestamp of the last scrape time",
        )
        for j in self.active_jobs:
            getattr(self, f"init_{j}")()

        # Start the server
        # Taken from https://github.com/prometheus/client_python/blob/master/prometheus_client/exposition.py#L163
        port = int(self.config.get("port", 8000))
        addr = self.config.get("addr", "0.0.0.0")

        class TmpServer(exposition.ThreadingWSGIServer):
            """Copy of ThreadingWSGIServer to update address_family locally"""

        TmpServer.address_family, addr = exposition._get_best_family(addr, port)
        app = exposition.make_wsgi_app(REGISTRY)
        # Hold onto the server class so that we can shut it down during cleanup
        self.httpd = exposition.make_server(
            addr, port, app, TmpServer, handler_class=exposition._SilentHandler
        )
        Thread(target=self.httpd.serve_forever, daemon=True).start()

    def init_item(self) -> None:
        """Initialize item metrics"""
        for s in self.item_stats:
            self.add_metric(
                s,
                Gauge,
                self.prefix + s.replace(".", "_"),
                f"rTorrent item stat {s}",
                [l.replace(".", "_") for l in self.item_labels],
            )

    def collect_item(self) -> None:
        """Collect item metrics"""
        call_list = self.item_stats + self.item_labels
        calls = [m + "=" for m in call_list]
        result = self.proxy.d.multicall2("", "main", *calls)
        for i in result:
            info = dict(list(zip(call_list, i)))
            for s in self.item_stats:
                self.metrics[s].labels(
                    **{k.replace(".", "_"): info[k] for k in self.item_labels}
                ).set(info[s])

    def init_tracker(self) -> None:
        """Initialize tracker metrics"""
        self.add_metric(
            "tracker_amount",
            Gauge,
            self.prefix + "tracker_amount",
            "Number of torrents belonging to a specific tracker",
            ["alias"],
        )
        self.add_metric(
            "tracker_errors",
            Gauge,
            self.prefix + "tracker_errors",
            "Number of torrents with tracker errors belonging to a specific tracker",
            ["alias"],
        )

    def collect_tracker(self) -> None:
        """Collect tracker metrics"""
        result = self.engine.items("main", ["d.custom=memo_alias", "d.message="])

        trackers = Counter([d.alias for d in result])
        tracker_errors = Counter([d.alias for d in result if d.message])

        for k, v in trackers.items():
            self.metrics["tracker_amount"].labels(alias=k).set(v)
            self.metrics["tracker_errors"].labels(alias=k).set(tracker_errors.get(k, 0))

    def init_system(self) -> None:
        """Initialize system metrics"""
        stat_methods = [
            "network.http.current_open",
            "network.http.max_open",
            "network.max_open_files",
            "network.max_open_sockets",
            "network.open_files",
            "network.open_sockets",
            "network.total_handshakes",
            "pieces.memory.block_count",
            "pieces.memory.current",
            "pieces.memory.max",
            "pieces.memory.sync_queue",
            "pieces.preload.min_rate",
            "pieces.preload.min_size",
            "pieces.preload.type",
            "pieces.stats.total_size",
            "pieces.stats_not_preloaded",
            "pieces.stats_preloaded",
            "pieces.sync.queue_size",
            "startup_time",
            "system.files.closed_counter",
            "system.files.failed_counter",
            "system.files.opened_counter",
            "throttle.global_down.max_rate",
            "throttle.global_down.rate",
            "throttle.global_down.total",
            "throttle.global_up.max_rate",
            "throttle.global_up.rate",
            "throttle.global_up.total",
            "throttle.max_downloads.global",
            "throttle.max_uploads.global",
            "throttle.unchoked_downloads",
            "throttle.unchoked_uploads",
        ]
        # Strip out unavailable methods
        self.system_stats = sorted(
            set(stat_methods) & set(self.proxy.system.listMethods())
        )
        # Set up metrics
        for name in self.system_stats:
            self.add_metric(
                name, Gauge, self.prefix + name.replace(".", "_"), f"rTorrent {name}"
            )
        self.add_metric(
            "view.size",
            Gauge,
            self.prefix + "view_size",
            "rTorrent view size",
            ["view"],
        )
        self.views = sorted(self.proxy.view.list())

    def collect_system(self) -> None:
        """Collect system metrics"""
        # Get data via multicall
        calls = [
            {"methodName": method, "params": []} for method in self.system_stats
        ] + [{"methodName": "view.size", "params": ["", view]} for view in self.views]

        result = [r[0] for r in self.proxy.system.multicall(calls)]

        # Get numeric metrics
        for m in self.system_stats:
            self.metrics[m].set(result[0])
            del result[0]

        result = result[-len(self.views) :]

        # Get view information
        for v in self.views:
            self.metrics["view.size"].labels(view=v).set(result[0])
            del result[0]

    def run(self):
        """Trigger all configured active collectors"""
        for j in self.active_jobs:
            getattr(self, f"collect_{j}")()
        self.metrics["scrape"].inc()
        self.metrics["last_scrape"].set(time())

    def cleanup(self):
        """Clean up metrics from the global registry and shut down the
        server"""
        for metric in self.metrics.values():
            REGISTRY.unregister(metric)
        self.httpd.shutdown()
        self.httpd.server_close()


if __name__ == "__main__":
    from pyrosimple import connect

    engine = connect()

    job = RtorrentExporter({"jobs": "system", "port": 8005})
    while True:
        job.run()
        sleep(5)



"""Torrent Item Filters.

Copyright (c) 2009, 2010, 2011 The PyroScope Project <pyroscope.project@gmail.com>

There's a lot of magic going on in this module, but essentially
its primary responsibility is parsing queries from rtcontrol and other
places. Once parsed, it uses the visitor pattern in some classes to do
things like return lists of the fields being referenced, prepare
d.multicall.filtered statements or check if it matches against an
actual item.
"""


import fnmatch
import math
import operator
import re
import time

from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Sequence, Union

from parsimonious.grammar import Grammar
from parsimonious.nodes import NodeVisitor

from pyrosimple import config, error, torrent
from pyrosimple.util.fmt import human_size


TRUE = {
    "true",
    "t",
    "yes",
    "y",
    "1",
}

FALSE = {
    "false",
    "f",
    "no",
    "n",
    "0",
}


@dataclass
class FilterOperator:
    """Small class to hold operator information"""

    name: str
    query_repr: str


Operators = {
    "eq": FilterOperator("eq", "="),
    "ne": FilterOperator("ne", "!="),
    "gt": FilterOperator("gt", ">"),
    "lt": FilterOperator("lt", "<"),
    "ge": FilterOperator("ge", ">="),
    "le": FilterOperator("le", "<="),
}


def truth(val: Any, context="statement") -> bool:
    """Convert truth value in "val" to a boolean."""
    # Try coercing it to an int then a bool
    try:
        return bool(0 + val)
    except TypeError:
        pass

    lower_val = val.lower()

    if lower_val in TRUE:
        return True
    if lower_val in FALSE:
        return False

    raise FilterError(
        f"Bad boolean value {val!r} in {context!r} (expected one of '{TRUE}', or '{FALSE}')"
    )


TIMEDELTA_UNITS = {
    "y": lambda d: d * 365 * 86400,
    "M": lambda d: d * 30 * 86400,
    "w": lambda d: d * 7 * 86400,
    "d": lambda d: d * 86400,
    "h": lambda d: d * 3600,
    "m": lambda d: d * 60,
    "s": lambda d: d,
}
TIMEDELTA_RE = re.compile(
    "^" + "".join(r"(?:(?P<{0}>\d+)[{0}{0}])?".format(i) for i in "yMwdhms") + "$"
)


def unquote_pre_filter(
    pre_filter: str, regex_: re.Pattern = re.compile(r"[\\]+")
) -> str:
    """Unquote a pre-filter condition."""
    if pre_filter.startswith('"') and pre_filter.endswith('"'):
        # Unquote outer level
        pre_filter = pre_filter[1:-1]
        pre_filter = regex_.sub(
            lambda x: x.group(0)[: len(x.group(0)) // 2], pre_filter
        )

    return pre_filter


class FilterError(error.UserError):
    """(Syntax) error in filter."""


class MatcherNode:
    """Base class for the tree structure."""

    def __init__(self, children: List):
        self.children = list(children)

    def match(self, item) -> bool:
        """Check if the item matches. All logic is deferred to subclasses."""
        raise NotImplementedError()

    def __repr__(self):
        result = type(self).__name__
        if self.children:
            result = str([str(c) for c in self.children])
        return result


class GroupNode(MatcherNode):
    """This simply groups another node, and optionally inverts it (a logical NOT)"""

    def __init__(self, children: List, invert: bool):
        super().__init__(children)
        self.invert = bool(invert)

    def match(self, item) -> bool:
        assert len(self.children) == 1
        result = bool(self.children[0].match(item))
        if self.invert:
            return not result
        return result

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        assert len(self.children) == 1
        inner = str(self.children[0].pre_filter())
        if self.invert:
            if inner.startswith('"not=$') and inner.endswith('"') and "\\" not in inner:
                return inner[6:-1]  # double negation, return inner command
            if inner.startswith('"'):
                inner = '"$' + inner[1:]
            else:
                inner = "$" + inner
            return "not=" + inner
        return inner

    def __repr__(self):
        prefix = "Not" if self.invert else ""
        return f"{prefix}{type(self).__name__}{[repr(c) for c in self.children]}"

    def to_match_string(self):
        """Turn the node to a valid query string"""
        ret_str = "! [ " if self.invert else "[ "
        child_strings = [c.to_match_string() for c in self.children]
        return ret_str + " ".join(child_strings) + " ]"


class AndNode(MatcherNode):
    """This node performs a logical AND for all of it's children."""

    def match(self, item) -> bool:
        return all(c.match(item) for c in self.children)

    def pre_filter(self):
        """Return rTorrent condition to speed up data transfer."""
        result = [x.pre_filter() for x in self.children]
        result = [x for x in result if x]
        if result:
            if int(config.settings.get("FAST_QUERY")) == 1 or len(result) == 1:
                return result[0]  # using just one simple expression is safer
            return f'and={",".join(result)}'
        return ""

    def to_match_string(self):
        """Turn the node to a valid query string"""
        if len(self.children) == 1:
            return self.children[0].to_match_string()
        child_strings = [c.to_match_string() for c in self.children]
        return " ".join(child_strings)

    def __repr__(self):
        return f"{type(self).__name__}{[repr(c) for c in self.children]}"


class OrNode(MatcherNode):
    """This node performs a logical OR for all of it's children."""

    def match(self, item) -> bool:
        return any(c.match(item) for c in self.children)

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        if len(self.children) == 1:
            return str(self.children[0].pre_filter())
        if int(config.settings.get("FAST_QUERY")) == 1:
            return ""
        result = [x.pre_filter() for x in self.children]
        result = [x for x in result if x]
        if result:
            return f'or={",".join(result)}'
        return ""

    def to_match_string(self):
        """Turn the node to a valid query string"""
        if len(self.children) == 1:
            return self.children[0].to_match_string()
        child_strings = [c.to_match_string() for c in self.children]
        return " OR ".join(child_strings)

    def __repr__(self):
        return f"{type(self).__name__}{[repr(c) for c in self.children]}"


class FieldFilter(MatcherNode):
    """Base class for all field filters.

    Subclasses of FieldFilter act as the leaves of the tree, providing
    matching and pre-filtering functionality."""

    def __init__(self, name: str, op: FilterOperator, value: str):
        """Store field name and filter value for later evaluations."""
        super().__init__([])  # Filters are the leaves of the tree and have no children
        self._name = name
        self._condition = self._value = value
        self._op: FilterOperator = op

        self.validate()

    def __str__(self) -> str:
        return str(self._name) + self._op.query_repr + str(self._condition)

    def __call__(self, item):
        return self.match(item)

    def validate(self):
        """Validate filter condition (template method)."""

    def match(self, item) -> bool:
        """Test if item matches filter.

        By default this will defer to the operator functions in subclasses,
        but that behaivor can be overridden."""
        return bool(getattr(self, self._op.name)(item))

    def eq(self, item) -> bool:
        """Test equality against item"""
        raise FilterError(
            f"Filter '{type(self).__name__}' for field '{self._name}' does not support comparison '{self._op}'"
        )

    def gt(self, item) -> bool:
        """Test if item is greater than value"""
        raise FilterError(
            f"Filter '{type(self).__name__}' for field '{self._name}' does not support comparison '{self._op}'"
        )

    # Unfortunate but necessary boilerplate functions
    # This makes it so that by default all operators are available to fields
    # If there's a more effecient way to implement it for a specific subclass, it needs
    # to happen in there.
    # pylint: disable=missing-function-docstring
    def ge(self, item) -> bool:
        return self.eq(item) or self.gt(item)

    def ne(self, item) -> bool:
        return not self.eq(item)

    def le(self, item) -> bool:
        return self.eq(item) or not self.gt(item)

    def lt(self, item) -> bool:
        return not self.eq(item) and not self.gt(item)

    # pylint: enable=missing-function-docstring

    def pre_filter(self) -> str:
        """Create a prefilter condition (if possible).

        By default this will defer to the operator functions in subclasses,
        but that behavior can be overridden."""
        method_name = f"pre_filter_{self._op.name}"
        return str(getattr(self, method_name)())

    def pre_filter_eq(self) -> str:
        """Returns empty if not defined in subclass."""
        return ""

    def pre_filter_gt(self) -> str:
        """Returns empty if not defined."""
        return ""

    def pre_filter_ne(self) -> str:
        """Create a prefilter by creating a NOT[name=filter] object and rendering it."""
        if self.pre_filter_eq():
            return GroupNode(
                [type(self)(self._name, Operators["eq"], self._value)], True
            ).pre_filter()
        return ""

    def to_match_string(self):
        """Turn the node to a valid query string"""
        return self._name + self._op.query_repr + str(self._value)


class PatternFilter(FieldFilter):
    """Pattern filter, either a glob or a /regex/ pattern."""

    CLEAN_PRE_VAL_RE = re.compile(r"(?:\[.*?\])|(?:\(.*?\))|(?:{.*?})|(?:\\)")
    SPLIT_PRE_VAL_RE = re.compile(r"[^a-zA-Z0-9/_]+")
    SPLIT_PRE_GLOB_RE = re.compile(r"[?*[\]]+")

    def validate(self) -> None:
        """Validate filter condition (template method)."""

        super().validate()
        self._value: str = self._value
        self._template = None
        self._flags = 0
        self._matcher: Callable[[str, Any], bool]
        if (
            self._value == '""'
        ):  # Replace an empty string with a simple truthiness check
            self._matcher = lambda val, _: val == ""
        elif self._value.startswith("/") and (
            self._value.endswith("/") or self._value.endswith("/i")
        ):
            # Pick out a couple regexes that can be simplified
            if self._value in ["//", "/.*/", "//i", "/.*/i"]:
                self._matcher = lambda _, __: True
            elif self._value in ["/.+/", "/.+/i"]:
                self._matcher = lambda val, __: bool(val)
            # Otherwise handle it generically
            else:
                value = self._value
                if self._value.endswith("/i"):
                    self._flags = re.IGNORECASE
                    value = self._value.rstrip("i")
                regex = re.compile(value[1:-1], self._flags)
                self._matcher = lambda val, _: bool(regex.search(val))
        elif self._value.startswith("{{") or self._value.endswith("}}"):
            self._template = self._value

            def _template_globber(val, item) -> bool:
                """Helper method to allow templating a glob."""
                if self._template is not None:
                    pattern = torrent.rtorrent.format_item(
                        torrent.rtorrent.env.from_string(self._template), item
                    )
                    return fnmatch.fnmatchcase(val, pattern)
                return False

            self._matcher = _template_globber
        else:
            # Pick out a glob that can be simplified
            if self._value == "*":
                self._matcher = lambda _, __: True
            else:
                self._matcher = (
                    lambda val, _: fnmatch.fnmatchcase(val, self._value)
                    or val == self._value
                )

    def pre_filter_eq(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        pf = prefilter_field_lookup(self._name)
        if pf is None or self._template:
            return ""
        if not self._value or self._value == '""':
            return f'"equal={pf},cat="'

        if self._value.startswith("/") and (
            self._value.endswith("/") or self._value.endswith("/i")
        ):
            if self._value.endswith("/i"):
                needle = self._value[1:-2]
            else:
                needle = self._value[1:-1]
            needle = self.CLEAN_PRE_VAL_RE.sub(" ", needle)
            split_needle = self.SPLIT_PRE_VAL_RE.split(needle)
            # If the cleaning of the needle did not succeed and we still
            # have regex-y values, don't prefilter
            if any(metachar in needle for metachar in r"{}[]\\|()"):
                return ""
        else:
            split_needle = self.SPLIT_PRE_GLOB_RE.split(self._value)
        # Grab the longest needle available from the array
        needle = list(sorted(split_needle, key=len))[-1]

        if needle:
            # Skip trying to filter on non-ASCII characters
            try:
                needle.encode("ascii")
            except UnicodeEncodeError:
                return ""
            needle = needle.replace('"', r"\\\"")
            return rf'"string.contains_i=${pf},\"{needle}\""'

        return ""

    def eq(self, item):
        """Return True if filter matches item."""
        val = getattr(item, self._name) or ""
        result = self._matcher(val, item)
        return result


class FilesFilter(PatternFilter):
    """Pattern filter on filenames in a torrent."""

    def match(self, item) -> bool:
        """Return True if filter matches item. Overridden from the
        parent class to deal with with an array of strings rather than
        a single string
        """
        val = getattr(item, self._name)
        if val is not None:
            for fileinfo in val:
                if fnmatch.fnmatchcase(fileinfo.path, self._value):
                    return True
            return False
        return False


class TaggedAsFilter(FieldFilter):
    """Case-insensitive tags filter. Tag fields are white-space
    separated lists of tags.
    """

    def pre_filter_eq(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        pf = prefilter_field_lookup(self._name)
        if pf is not None:
            if self._exact and not self._value:
                return f'"equal={pf},cat="'
            if not self._value:
                return f'"not=${pf}"'
            needle = self._value.replace('"', r"\\\"")
            return rf'"string.contains_i=${pf},\"{needle}\""'
        return ""

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()
        self._value = self._value.lower()

        # If the tag starts with ':', test for exact equality (just this tag, no others)
        if self._value.startswith(":"):
            self._exact = True
            self._value = self._value[1:]
        else:
            self._exact = not self._value

    def eq(self, item) -> bool:
        """Return True if filter matches item."""
        tags = getattr(item, self._name) or []
        if self._exact:
            # Exact equality check
            return set(self._value) == set(tags)
        # Is given tag in list?
        return self._value in tags


class BoolFilter(FieldFilter):
    """Filter boolean values."""

    def pre_filter_eq(self):
        """Return rTorrent condition to speed up data transfer."""
        pf = prefilter_field_lookup(self._name)
        if pf is not None:
            return f'"equal={pf},value={int(self._value)}"'
        return ""

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()

        self._value = truth(str(self._value), self._condition)
        self._condition = "yes" if self._value else "no"

    def eq(self, item):
        """Return True if filter matches item."""
        val = getattr(item, self._name) or False
        return bool(val) is self._value

    def to_match_string(self):
        """Turn the node to a valid query string"""
        val = "yes" if self._value else "no"
        return self._name + self._op.query_repr + val


class NumericFilterBase(FieldFilter):
    """Base class for numerical value filters."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.not_null = False

    def match(self, item) -> bool:
        """Return True if filter matches item."""
        val = getattr(item, self._name) or 0
        if self.not_null and self._value and not val:
            return False
        # Grab the function from the native operator module
        op = getattr(operator, self._op.name)
        return bool(op(float(val), float(self._value)))


def prefilter_field_lookup(name: str) -> Optional[str]:
    """Return the prefield field for a given name (if available)"""
    field = torrent.engine.field_lookup(name)
    if field is None:
        return None
    return field.prefilter_field


class FloatFilter(NumericFilterBase):
    """Filter float values."""

    FIELD_SCALE = {
        "ratio": 1000,
    }

    def pre_filter(self):
        """Prefilter a float value.

        rTorrent doesn't actually have floats, so we need to do a
        little translation for the prefiltering
        """
        pf = prefilter_field_lookup(self._name)
        if pf is None:
            return ""
        val = int(self._value) * self.FIELD_SCALE.get(self._name, 1)
        lookup_table = {
            "eq": ("equal", int(val)),
            "ge": ("greater", math.floor(val - 1)),
            "gt": ("greater", math.floor(val)),
            "le": ("less", math.ceil(val + 1)),
            "lt": ("less", math.ceil(val)),
        }
        op, val = lookup_table[self._op.name]
        return f'"{op}=value=${pf},value={val}"'

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()

        try:
            self._value = float(self._value)
        except (ValueError, TypeError) as exc:
            raise FilterError(
                f"Bad numerical value {self._value!r} in {self._condition!r} ({exc})"
            ) from exc


class TimeFilter(NumericFilterBase):
    """Filter UNIX timestamp values."""

    TIMEDELTA_UNITS = {
        "y": lambda d: d * 365 * 86400,
        "M": lambda d: d * 30 * 86400,
        "w": lambda d: d * 7 * 86400,
        "d": lambda d: d * 86400,
        "h": lambda d: d * 3600,
        "m": lambda d: d * 60,
        "s": lambda d: d,
    }
    TIMEDELTA_RE = re.compile(
        "^" + "".join(r"(?:(?P<{0}>\d+)[{0}{0}])?".format(i) for i in "yMwdhms") + "$"
    )

    def __init__(self, name: str, op: FilterOperator, value: str):
        # During validate(), one of these two must be set to something
        # non-None
        self._timestamp_offset = None
        self._timestamp = None
        super().__init__(name, op, value)
        self.not_null = False

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        # A "0" might indicate just that, or possibly an empty
        # custom value.
        if self._value == 0:
            return ""
        pf = prefilter_field_lookup(self._name)
        # Add a day of wiggle room to avoid any possible timezone problems
        time_fuzz = 60 * 60 * 24
        timestamp = 0
        cmp_ = ""
        if pf is None:
            return ""
        if self._op.name in ["gt", "ge"]:
            timestamp = int(float(self._value)) - time_fuzz
            cmp_ = "greater"
        elif self._op.name in ["lt", "le"]:
            timestamp = int(float(self._value)) + time_fuzz
            cmp_ = "less"

        if timestamp and cmp_:
            return f'"{cmp_}=value=${pf},value={int(timestamp)}"'
        return ""

    def validate(self):
        # 0 is a special case
        delta = self._parse_delta()
        if delta is not None:
            self._timestamp_offset = delta
            # Invert the operators to be more intuitive, i.e. <3d
            # should match for values of less than 3 days *ago*.
            self._invert_operator()
            return
        self._timestamp = self._parse_absolute_timestamp()
        if self._timestamp is None:
            raise ValueError(f"Could not parse timestamp {self._condition!r}")

    @property
    def _value(self) -> str:
        if self._timestamp_offset is not None:
            return str(time.time() - self._timestamp_offset)
        if self._timestamp is not None:
            return str(self._timestamp)
        raise ValueError(f"Unset time value from condition {self._condition!r}")

    @_value.setter
    def _value(self, _: str) -> None:
        """Discard attempts to set the value, primarily because the
        grandparent FieldFilter tries to do it in __init__()"""
        return None

    def _invert_operator(self):
        if self._op.name == "gt":
            self._op = Operators["le"]
        elif self._op.name == "ge":
            self._op = Operators["lt"]
        elif self._op.name == "lt":
            self._op = Operators["ge"]
        elif self._op.name == "le":
            self._op = Operators["gt"]

    def _parse_delta(self) -> Optional[int]:
        match = self.TIMEDELTA_RE.match(self._condition)
        if not match:
            return None
        delta_val = 0
        for unit, val in match.groupdict().items():
            if val:
                delta_val = self.TIMEDELTA_UNITS[unit](int(val, 10))
        return delta_val

    def _parse_absolute_timestamp(self) -> Optional[int]:
        if str(self._condition).isdigit():
            # Literal UNIX timestamp
            try:
                return int(self._condition)
            except (ValueError, TypeError) as exc:
                raise FilterError(
                    f"Bad UNIX timestamp value {self._condition!r}"
                ) from exc
        # Assume it's an absolute date
        if "/" in self._condition:
            # U.S.
            dtfmt = "%m/%d/%Y"
        elif "." in self._condition:
            # European
            dtfmt = "%d.%m.%Y"
        else:
            # Fall back to ISO
            dtfmt = "%Y-%m-%d"

        val = str(self._condition).upper().replace(" ", "T")
        if "T" in val:
            # Time also given
            dtfmt += "T%H:%M:%S"[: 3 + 3 * val.count(":")]

        try:
            timestamp = time.mktime(time.strptime(val, dtfmt))
        except ValueError as exc:
            raise FilterError(
                f"Could not parse timestamp {self._condition!r} with format {dtfmt!r}"
            ) from exc
        return int(timestamp)


class TimeFilterNotNull(TimeFilter):
    """Filter UNIX timestamp values, ignore unset values unless compared to 0."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.not_null = True


class DurationFilter(TimeFilter):
    """Filter durations in seconds."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.not_null = True

    @property
    def _value(self):
        if self._condition == "0":
            return 0
        if self._timestamp_offset is not None:
            return self._timestamp_offset
        if self._timestamp is not None:
            return time.time() - self._timestamp
        raise ValueError(f"Unset time value from condition {self._condition!r}")

    @_value.setter
    def _value(self, _):
        """Discard attempts to set the value, primarily because the
        grandparent FieldFilter tries to do it in __init__()"""
        return None

    def validate(self):
        # 0 is a special case
        if self._condition == "0":
            return
        delta = self._parse_delta()
        if delta is not None:
            self._timestamp_offset = delta
            return
        self._timestamp = self._parse_absolute_timestamp()
        if self._timestamp is None:
            raise ValueError(f"Could not parse timestamp {self._condition!r}")

    def match(self, item) -> bool:
        """Return True if filter matches item."""
        if getattr(item, self._name) is None:
            # Never match "N/A" items, except when "-0" was specified
            return not bool(self._value)
        return super().match(item)

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        # A "0" might indicate just that, or possibly an empty
        # custom value.
        if self._value == 0:
            return ""
        pf = prefilter_field_lookup(self._name)
        time_fuzz = (
            60 * 60 * 24
        )  # Add a day of wiggle room to avoid any possible timezone problems
        timestamp = 0
        cmp_ = ""
        if pf is None:
            return ""
        if self._op.name in ["gt", "ge"]:
            timestamp = self._value + time_fuzz
            cmp_ = "greater"
        elif self._op.name in ["lt", "le"]:
            timestamp = self._value - time_fuzz
            cmp_ = "less"

        if timestamp and cmp_:
            return f'"{cmp_}=value=${pf},value={int(timestamp)}"'
        return ""


class ByteSizeFilter(NumericFilterBase):
    """Filter size and bandwidth values."""

    UNITS = {"b": 1, "k": 1024, "m": 1024**2, "g": 1024**3, "t": 1024**4}

    def pre_filter(self) -> str:
        """Return rTorrent condition to speed up data transfer."""
        comparers = {
            "gt": "greater",
            "lt": "less",
            "eq": "equal",
        }
        pf = prefilter_field_lookup(self._name)
        if pf is not None and self._op.name in comparers:
            return '"{}={},value={}"'.format(
                comparers[self._op.name],
                pf,
                int(self._value),
            )
        return ""

    def validate(self):
        """Validate filter condition (template method)."""
        super().validate()

        # Get scale
        lower_val = str(self._value).lower()
        if any(lower_val.endswith(i) for i in self.UNITS):
            scale = self.UNITS[lower_val[-1]]
            self._value = self._value[:-1]
        else:
            scale = 1

        # Get float value
        try:
            self._value = float(self._value)
        except (ValueError, TypeError) as exc:
            raise FilterError(
                f"Bad numerical value {self._value!r} in {self._condition!r}"
            ) from exc

        # Scale to bytes
        self._value = self._value * scale

    def to_match_string(self):
        """Turn the node to a valid query string"""
        return (
            self._name
            + self._op.query_repr
            + human_size(self._value).replace(" ", "").replace("iB", "")
        )


QueryGrammar = Grammar(
    r"""
    query = (group / stmt) (ws (group / stmt))*
    stmt = (or_stmt / conds)
    or_stmt = (group / conds) (ws or ws (group / conds))*
    group = (not ws)? lpar ws stmt ws rpar
    conds = cond (ws cond)*
    cond = (&or / &lpar / &rpar / &not / named_cond / unnamed_cond)
    named_cond = word conditional filter
    unnamed_cond = filter
    filter      = (quoted_regex / quoted / glob / regex / word)
    glob = ~r"[\S]+"
    quoted_regex =  ~r'"/[^/\"]*/i?"'
    regex = ~r"/[^/]*/i?"
    quoted      = ~'"[^\"]*"'
    word        = ~r"[\w]+"
    conditional = (ne / ge / le / lt / gt / eq)
    ne          = ("!=" / "<>")
    eq          = ("==" / "=")
    gt          = (">" / "=+")
    lt          = ("<")
    ge          = (">=")
    le          = ("<=" / "=-")
    fws         = ~r"\s+"
    ws          = ~r"\s*"
    lpar = "["
    rpar = "]"
    or   = "OR"
    not  = ( "NOT" / "!" )
    """
)


class KeyNameVisitor(NodeVisitor):
    """Walk through a query tree and returns an array of key names.
    Implicit key names are not included."""

    # pylint: disable=unused-argument,missing-function-docstring
    def visit_expr(self, node, visited_children):
        output = ""
        for child in visited_children:
            output += str(child)
        return output

    def visit_named_cond(self, node, visited_children):
        return visited_children[0]

    def visit_word(self, node, visited_children):
        return [node.text]

    def visit_filter(self, node, visited_children):
        return ["name"]

    def generic_visit(self, node, visited_children):
        if visited_children:
            return [item for sublist in visited_children for item in sublist]
        return []


def create_filter(name: str, op: str, value: str):
    """Generates a filter class with the given name, operation and value"""
    field = torrent.engine.field_lookup(name)
    if field is None:
        raise SyntaxError(f"No such field '{name}'")
    filt = field._matcher
    return filt(name, op, value)


class MatcherBuilder(NodeVisitor):
    """Build a simplified tree of MatcherNodes to match an item against."""

    # pylint: disable=unused-argument,missing-function-docstring
    def visit_named_cond(self, node, visited_children):
        key, cond, needle = visited_children
        return create_filter(key, cond, needle)

    def visit_group(self, node, visited_children):
        real_children = [c for c in visited_children[1:] if c is not None]
        return GroupNode(
            real_children, visited_children[0]
        )

    def visit_not(self, node, visited_children):
        if node.text:
            return True
        return False

    def __pare_children(self, children, class_):
        """Get all non-None children, and if there's only one child left,
        return the child instead of wrapping it in the parent class.
        """
        real_children = [c for c in children if c is not None]
        if len(real_children) == 0:
            return None
        if len(real_children) == 1:
            return real_children[0]
        return class_(real_children)

    def visit_or_stmt(self, node, visited_children):
        children = [visited_children[0]]
        if visited_children[1] is not None:
            if isinstance(visited_children[1], list):
                for c in visited_children[1]:
                    children.append(c)
            else:
                children.append(visited_children[1])
        if len(children) == 1:
            return children[0]
        return OrNode(children)

    def visit_conds(self, node, visited_children):
        if len(visited_children) == 2 and isinstance(visited_children[1], list):
            children = [visited_children[0]] + visited_children[1]
        else:
            children = visited_children
        pared = self.__pare_children(children, AndNode)
        return pared

    def visit_cond(self, node, visited_children):
        if len(visited_children) == 1:
            child = visited_children[0]
            if isinstance(child, str):
                return create_filter("name", Operators["eq"], child)
            if isinstance(child, re.Pattern):
                return create_filter("name", Operators["eq"], child.pattern)
        return self.generic_visit(node, visited_children)

    def visit_query(self, node, visited_children):
        real_children = [c for c in visited_children if c is not None]
        if len(real_children) == 1:
            return real_children[0]
        elif isinstance(real_children[1], list):
            return AndNode([real_children[0]] + real_children[1])
        else:
            return AndNode([real_children[0]] + real_children[1:])

    # Unfortunate but necessary boilerplate methods

    def visit_word(self, node, visited_children):
        return node.text

    def visit_quoted(self, node, visited_children):
        return node.text[1:-1]

    def visit_glob(self, node, visited_children):
        return node.text

    def visit_regex(self, node, visited_children):
        return node.text

    def visit_quoted_regex(self, node, visited_children):
        return node.text[1:-1]

    def visit_eq(self, node, visited_children):
        return Operators["eq"]

    def visit_ne(self, node, visited_children):
        return Operators["ne"]

    def visit_gt(self, node, visited_children):
        return Operators["gt"]

    def visit_ge(self, node, visited_children):
        return Operators["ge"]

    def visit_lt(self, node, visited_children):
        return Operators["lt"]

    def visit_le(self, node, visited_children):
        return Operators["le"]

    def generic_visit(self, node, visited_children):
        real_children = [c for c in visited_children if c is not None]
        if real_children:
            if isinstance(real_children, list) and len(real_children) == 1:
                return real_children[0]
            return real_children
        return None


RE_UNQUOTED_REGEX = re.compile(r"(\w*)(!?=)(\/.*\/)")


def cli_args_to_match_str(query: Sequence) -> str:
    """Convert CLI arguments to a string. Most usefully, this will
    automatically double quote unnamed conditions or regexes if they
    have a space in them.
    """
    args = []
    for a in query:
        if " " in a and not set("=><") & set(a) and not a.startswith('"'):
            a = f'"{a}"'
        if " " in a and RE_UNQUOTED_REGEX.match(a):
            a = RE_UNQUOTED_REGEX.sub(r'\1\2"\3"', a)
        args.append(a)
    return " ".join(args)


def create_matcher(query: Union[str, Sequence[str]]):
    """Utility function to build a matcher from a query string."""
    if not isinstance(query, str):
        query_str = cli_args_to_match_str(query)
    else:
        query_str = query
    return MatcherBuilder().visit(QueryGrammar.parse(query_str))



""" Perform raw XMLRPC calls.

    Copyright (c) 2010 The PyroScope Project <pyroscope.project@gmail.com>
"""


import difflib
import json
import logging
import os
import shlex
import sys
import tempfile

from pprint import pformat
from xmlrpc import client as xmlrpc_client

from pyrosimple import config, error
from pyrosimple.scripts.base import ScriptBaseWithConfig
from pyrosimple.util import fmt, rpc


def read_blob(arg: str) -> bytes:
    """Read a BLOB from given ``@arg``."""
    if arg == "@-":
        return sys.stdin.buffer.read()

    if any(arg.startswith(f"@{x}://") for x in ["http", "https", "ftp", "file"]):
        try:
            import requests  # pylint: disable=import-outside-toplevel

            response = requests.get(arg[1:], timeout=60)
            response.raise_for_status()
            return response.content
        except requests.RequestException as exc:
            raise error.UserError(str(exc))

    with open(os.path.expanduser(arg[1:]), "rb") as handle:
        return handle.read()


class RtorrentXmlRpc(ScriptBaseWithConfig):
    """
    Perform raw rTorrent RPC calls, like "rtxmlrpc throttle.global_up.max_rate".

    Start arguments with "+" or "-" to indicate they're numbers (type i4 or i8).
    Use "[1,2,..." for arrays. Use "@" to indicate binary data, which can be
    followed by a file path (e.g. "@/path/to/file"), a URL (https, http, ftp,
    and file are supported), or '-' to read from stdin.
    """

    def __init__(self):
        super().__init__()
        self.proxies = []

    def add_options(self):
        """Add program options."""
        super().add_options()

        # basic options
        self.parser.add_argument("method", nargs="?", help="method name")
        self.parser.add_argument("args", nargs="*", help="arguments to method")
        self.parser.add_argument(
            "-o",
            "--output-format",
            default="pretty",
            choices=["pretty", "repr", "json"],
            help="Output format to use. Defaults to 'pretty'",
        )
        self.add_bool_option(
            "-i",
            "--as-import",
            help="execute each argument as a private command using 'import'",
        )
        self.add_bool_option(
            "--session",
            "--restore",
            help="restore session state from .rtorrent session file(s)",
        )
        self.add_bool_option(
            "--repl",
            help="Open an interactive prompt to run commands",
        )

    def open(self):
        """Open connections and return proxies."""
        if not self.proxies:
            if not config.settings["SCGI_URL"]:
                config.autoload_scgi_url()
            if not config.settings["SCGI_URL"]:
                self.log.error(
                    "Unable to automatically detect a RPC connection, see"
                    " https://kannibalox.github.io/pyrosimple/configuration/#reference"
                )
            for url in config.multi_connection_lookup(
                self.options.url or config.settings["SCGI_URL"]
            ):
                if url:
                    self.proxies.append(rpc.RTorrentProxy(url))
        return self.proxies

    def cooked(self, raw_args):
        """Return interpreted / typed list of args."""
        args = []
        for arg in raw_args:
            if arg and arg[0] in "+-":
                try:
                    arg = int(arg, 10)
                except (ValueError, TypeError) as exc:
                    self.log.warning("Not a valid number: %r (%s)", arg, exc)
            elif arg.startswith("[["):  # escaping, not a list
                arg = arg[1:]
            elif arg == "[]":
                arg = []
            elif arg.startswith("["):
                arg = arg[1:].split(",")
                if all(i.isdigit() for i in arg):
                    arg = [int(i, 10) for i in arg]
            elif arg.startswith("@"):
                arg = xmlrpc_client.Binary(read_blob(arg))
            args.append(arg)

        return args

    def execute(self, proxy, method, args):
        """Execute given RPC call."""
        try:
            result = getattr(proxy, method)(*tuple(args))
        except rpc.ERRORS as exc:
            self.log.error(
                "While calling %s(%s) via %s: %s",
                method,
                ", ".join(repr(i) for i in args),
                proxy,
                exc,
            )
            if f"Method '{method}' not defined" in str(
                exc
            ) or f"method not found: {method}" in str(exc):
                cmds = difflib.get_close_matches(
                    method, proxy.system.listMethods() + ["system.listMethods"]
                )
                if cmds:
                    print("The most similar methods are:")
                    for w in cmds:
                        print("-", w)
                else:
                    print(
                        "No similar methods found, try `rtxmlrpc system.listMethods` to see a full list of available methods."
                    )
            if isinstance(exc, rpc.HashNotFound):
                self.return_code = error.EX_NOINPUT
            else:
                self.return_code = error.EX_DATAERR
        else:
            if self.log.isEnabledFor(
                logging.WARNING
            ):  # Hack to hide output when `-q` is in effect
                if self.options.output_format == "repr":
                    result = pformat(result)
                elif self.options.output_format == "json":
                    result = json.dumps(result)
                else:
                    result = fmt.rpc_result_to_string(result)
                print(result)

    def do_import(self):
        """Handle import files or streams passed with '-i'."""
        tmp_import = None
        # Concatenate arguments
        if self.args is None:
            self.args = [self.options.method]
        else:
            self.args = [self.options.method] + self.args
        try:
            if self.args[0].startswith("@") and self.args[0] != "@-":
                import_file = os.path.expanduser(self.args[0][1:])
                if not os.path.isfile(import_file):
                    self.parser.error(f"File not found (or not a file): {import_file}")
                args = (rpc.NOHASH, os.path.abspath(import_file))
            else:
                script_text = "\n".join(self.args + [""])
                if script_text == "@-\n":
                    script_text = sys.stdin.read()

                with tempfile.NamedTemporaryFile(
                    suffix=".rc", prefix="rtxmlrpc-", delete=False
                ) as handle:
                    handle.write(script_text.encode("utf-8"))
                    tmp_import = handle.name
                args = (rpc.NOHASH, tmp_import)

            for proxy in self.open():
                self.execute(proxy, "import", args)
        finally:
            if tmp_import and os.path.exists(tmp_import):
                os.remove(tmp_import)

    def do_command(self):
        """Call a single command with arguments."""
        if not self.options.method:
            self.parser.print_help()
            print(
                "No method name given! Try `rtxmlrpc system.listMethods` to see a list of available methods."
            )
            sys.exit(error.EX_USAGE)
        method = self.options.method

        raw_args = self.args
        if "=" in method:
            if raw_args:
                self.parser.error(
                    "Please don't mix rTorrent and shell argument styles!"
                )
            method, raw_args = method.split("=", 1)
            raw_args = raw_args.split(",")

        for proxy in self.open():
            self.execute(proxy, method, self.cooked(raw_args))

    def print_repl_help(self):
        """Short REPL help output"""
        import textwrap  # pylint: disable=import-outside-toplevel

        print(
            textwrap.dedent(
                r"""\
        Entering prompt. Press Ctrl-D to exit.
        rTorrent XMLRPC REPL Help Summary
        =================================

        <Ctrl-d>            Exit the REPL.
        <Ctrl-c>            Cancel the current command.
        ?                   Show this help text.
        \help               Show this help text.
        \stats              Show current call stats.
        \connect URL        Connect to a different host
        cmd=arg1,arg2,..    Call a XMLRPC command"""
            )
        )

    def do_repl(self):
        """Run a simple REPL loop"""
        self.open()
        # pylint: disable=import-outside-toplevel
        from prompt_toolkit import PromptSession
        from prompt_toolkit.completion import WordCompleter

        # pylint: enable=import-outside-toplevel

        session = PromptSession(
            completer=WordCompleter(
                self.proxies[0].system.listMethods()
                + [r"\help", r"\stats", r"\connect"],
                WORD=True,
            )
        )
        self.print_repl_help()
        while True:
            try:
                label = ",".join([p.system.hostname() for p in self.proxies])
                text = session.prompt(f"{label}> ")
                if not text:
                    continue
                if text in {"?", r"\help"}:
                    self.print_repl_help()
                    continue
                if text == r"\stats":
                    print(self.rpc_stats())
                    continue
                if text.startswith(r"\connect "):
                    config.settings["SCGI_URL"] = text.split(" ")[1]
                    self.proxies = []
                    self.open()
                    continue
                self.options.method, *self.args = shlex.split(text)
                self.do_command()
            except KeyboardInterrupt:
                continue  # Control-C pressed. Try again.
            except EOFError:
                break  # Control-D pressed.

    def mainloop(self):
        """The main loop."""
        # Dispatch to handlers
        if self.options.as_import:
            self.do_import()
        elif self.options.repl:
            self.do_repl()
        else:
            self.do_command()

        # RPC stats
        self.log.debug("RPC stats: %s", self.rpc_stats())


def run():  # pragma: no cover
    """The entry point."""
    RtorrentXmlRpc().run()


if __name__ == "__main__":
    run()
